import os
from functools import partial
from datetime import datetime
from PySide6.QtCore import Qt, QThreadPool, QFileSystemWatcher
from PySide6.QtGui import QFont
from PySide6.QtSvgWidgets import QSvgWidget
from PySide6.QtWidgets import (
    QTabWidget,
    QWidget,
    QGridLayout,
    QLabel,
    QHeaderView,
    QPushButton,
    QSizePolicy,
    QHBoxLayout,
    QSpacerItem,
    QGroupBox,
    QVBoxLayout,
    QMessageBox,
    QListWidget,
    QFileDialog,
    QTreeWidget,
    QErrorMessage,
    QFileSystemModel,
    QTreeView,
    QComboBox,
)

from swane import strings
from swane.config.config_enums import GlobalPrefCategoryList
from swane.workers.SlicerExportWorker import SlicerExportWorker
from swane.workers.SlicerViewerWorker import SlicerViewerWorker
from swane.ui.CustomTreeWidgetItem import CustomTreeWidgetItem
from swane.ui.PersistentProgressDialog import PersistentProgressDialog
from swane.ui.PreferencesWindow import PreferencesWindow
from swane.ui.VerticalScrollArea import VerticalScrollArea
from swane.config.ConfigManager import ConfigManager
from swane.workers.DicomSearchWorker import DicomSearchWorker
from swane.utils.DataInputList import DataInputList
from swane.utils.DependencyManager import DependencyManager
from swane.config.preference_list import WORKFLOW_TYPES
from swane.nipype_pipeline.engine.WorkflowReport import WorkflowReport, WorkflowSignals
from swane.utils.Subject import Subject, SubjectRet
from swane.workers.open_results_directory import open_results_directory


class SubjectTab(QTabWidget):
    """
    Custom implementation of PySide QTabWidget to define a subject tab widget.

    """

    DATATAB = 0
    EXECTAB = 1
    RESULTTAB = 2

    def __init__(
        self, global_config: ConfigManager, subject: Subject, main_window, parent=None
    ):
        super(SubjectTab, self).__init__(parent)
        self.global_config = global_config
        self.subject = subject
        self.main_window = main_window

        self.data_tab = QWidget()
        self.exec_tab = QWidget()
        self.result_tab = QWidget()

        self.addTab(self.data_tab, strings.subj_tab_data_tab_name)
        self.addTab(self.exec_tab, strings.subj_tab_wf_tab_name)
        self.addTab(self.result_tab, strings.subj_tab_results_tab_name)

        self.directory_watcher = QFileSystemWatcher()
        self.directory_watcher.directoryChanged.connect(self.reset_workflow)

        self.scan_directory_watcher = QFileSystemWatcher()
        self.scan_directory_watcher.directoryChanged.connect(self.clear_scan_result)

        self.result_directory_watcher = QFileSystemWatcher()
        self.result_directory_watcher.directoryChanged.connect(
            self.enable_tab_if_result_dir
        )
        self.result_directory_watcher.addPath(self.subject.folder)

        self.workflow_process = None
        self.node_list = None
        self.input_report = {}
        self.dicom_scan_series_list = []
        self.importable_series_list = QListWidget()
        self.workflow_type_combo = None
        self.generate_workflow_button = None
        self.node_list_treeWidget = None
        self.subject_config_button = None
        self.exec_button = None
        self.exec_graph = None
        self.load_scene_button = None
        self.open_results_directory_button = None
        self.results_model = None
        self.result_tree = None
        self.generate_scene_button = None
        self.workflow_had_error = False

        self.data_tab_ui()
        self.exec_tab_ui()
        self.result_tab_ui()

        self.setTabEnabled(SubjectTab.EXECTAB, False)
        self.setTabEnabled(SubjectTab.RESULTTAB, False)

    def update_node_list(self, wf_report: WorkflowReport):
        """
        Searches for the node linked to the wf_report arg.
        Uses the wf_report arg to update the node status.

        Parameters
        ----------
        wf_report : WorkflowReport
            Workflow Monitor Worker message to parse.

        Returns
        -------
        None.

        """

        if wf_report.signal_type == WorkflowSignals.WORKFLOW_STOP:
            self.enable_tab_if_result_dir(on_workflow_running=True)
            self.setTabEnabled(SubjectTab.DATATAB, True)
            self.exec_button_set_enabled(False)

            if self.workflow_had_error:
                self.generate_workflow_button.setEnabled(True)
                self.subject.workflow = None
                self.exec_button.setText(strings.subj_tab_wf_executed_with_error)
                self.exec_button.setToolTip("")
            else:
                self.exec_button.setText(strings.subj_tab_wf_executed)
                self.exec_button.setToolTip("")

            shutdown = self.global_config.getboolean_safe(
                GlobalPrefCategoryList.MAIN, "shutdown"
            )
            if shutdown:
                self.main_window.safe_shutdown_after_workflow(self.subject)

            return
        elif wf_report.signal_type == WorkflowSignals.INVALID_SIGNAL:
            # Invalid signal sent from WF to UI, code error intercept
            try:
                self.workflow_process.stop_event.set()
            except:
                pass
            msg_box = QMessageBox()
            msg_box.setText(strings.subj_tab_wf_invalid_signal)
            msg_box.exec()

        # TODO - To be implemented for RAM usage info by each workflow
        # if msg == WorkflowProcess.WORKFLOW_INSUFFICIENT_RESOURCES:
        #     msg_box = QMessageBox()
        #     msg_box.setText(strings.pttab_wf_insufficient_resources)
        #     msg_box.exec()

        if wf_report.signal_type == WorkflowSignals.NODE_STARTED:
            icon = self.main_window.LOADING_MOVIE_FILE
        elif wf_report.signal_type == WorkflowSignals.NODE_COMPLETED:
            icon = self.main_window.OK_ICON_FILE
        else:
            icon = self.main_window.ERROR_ICON_FILE
            self.workflow_had_error = True
            # Mail manager initialization
            mail_manager = self.global_config.get_mail_manager()
            if mail_manager is not None:
                try:
                    mail_manager.send_report(
                        f"{self.subject.name} - {wf_report.workflow_name} - {wf_report.node_name} FAILED at {datetime.now()}"
                    )
                except:
                    pass

        self.node_list[wf_report.workflow_name].node_list[
            wf_report.node_name
        ].node_holder.set_art(icon)

        if wf_report.info is not None:
            self.node_list[wf_report.workflow_name].node_list[
                wf_report.node_name
            ].node_holder.setToolTip(0, wf_report.info)

        self.node_list[wf_report.workflow_name].node_holder.setExpanded(True)

        if icon == self.main_window.OK_ICON_FILE:
            completed = True
            for key in self.node_list[wf_report.workflow_name].node_list.keys():
                if (
                    self.node_list[wf_report.workflow_name]
                    .node_list[key]
                    .node_holder.art
                    != self.main_window.OK_ICON_FILE
                ):
                    completed = False
                    break
            if completed:
                self.node_list[wf_report.workflow_name].node_holder.set_art(
                    self.main_window.OK_ICON_FILE
                )
                self.node_list[wf_report.workflow_name].node_holder.setExpanded(False)
                self.node_list[wf_report.workflow_name].node_holder.completed = True
                # Mail manager initialization
                mail_manager = self.global_config.get_mail_manager()
                if mail_manager is not None:
                    try:
                        mail_manager.send_report(
                            f"{self.subject.name} - {wf_report.workflow_name} COMPLETED at {datetime.now()}"
                        )
                    except:
                        pass

    def remove_running_icon(self):
        """
        Remove all the loading icons from the series labels.

        Returns
        -------
        None.

        """

        for key1 in self.node_list.keys():
            for key2 in self.node_list[key1].node_list.keys():
                if (
                    self.node_list[key1].node_list[key2].node_holder.art
                    == self.main_window.LOADING_MOVIE_FILE
                ):
                    self.node_list[key1].node_list[key2].node_holder.set_art(
                        self.main_window.VOID_SVG_FILE
                    )

    def data_tab_ui(self):
        """
        Generates the Data tab UI.

        Returns
        -------
        None.

        """

        # Horizontal Layout
        layout = QHBoxLayout()

        # First Column: INPUT LIST
        scroll_area = VerticalScrollArea()
        folder_layout = QGridLayout()
        scroll_area.m_scrollAreaWidgetContents.setLayout(folder_layout)

        bold_font = QFont()
        bold_font.setBold(True)
        x = 0

        for data_input in self.subject.input_state_list:
            self.input_report[data_input] = [
                QSvgWidget(self),
                QLabel(data_input.value.label),
                QLabel(""),
                QPushButton(strings.subj_tab_import_button),
                QPushButton(strings.subj_tab_clear_button),
                None,
            ]
            self.set_error(data_input, "")
            if data_input.value.tooltip != "":
                # Add tooltips and append ⓘ character to label
                self.input_report[data_input][1].setText(
                    data_input.value.label + " " + strings.INFOCHAR
                )
                self.input_report[data_input][1].setToolTip(data_input.value.tooltip)
            self.input_report[data_input][1].setFont(bold_font)
            self.input_report[data_input][1].setAlignment(Qt.AlignLeft | Qt.AlignBottom)
            self.input_report[data_input][2].setAlignment(Qt.AlignLeft | Qt.AlignTop)
            self.input_report[data_input][2].setStyleSheet("margin-bottom: 20px")
            self.input_report[data_input][3].clicked.connect(
                lambda checked=None, z=data_input: self.dicom_import_to_folder(z)
            )
            self.input_report[data_input][3].setSizePolicy(
                QSizePolicy.Fixed, QSizePolicy.Fixed
            )
            self.input_report[data_input][4].clicked.connect(
                lambda checked=None, z=data_input: self.clear_import_folder(z)
            )
            self.input_report[data_input][4].setSizePolicy(
                QSizePolicy.Fixed, QSizePolicy.Fixed
            )

            folder_layout.addWidget(self.input_report[data_input][0], (x * 2), 0, 2, 1)
            folder_layout.addWidget(self.input_report[data_input][1], (x * 2), 1)

            folder_layout.addWidget(self.input_report[data_input][3], (x * 2), 2)
            folder_layout.addWidget(self.input_report[data_input][4], (x * 2), 3)

            folder_layout.addWidget(
                self.input_report[data_input][2], (x * 2) + 1, 1, 1, 3
            )
            x += 1

        # Second Column: Series to be imported
        import_group_box = QGroupBox()
        import_layout = QVBoxLayout()
        import_group_box.setLayout(import_layout)

        scan_dicom_folder_button = QPushButton(strings.subj_tab_scan_dicom_button)
        scan_dicom_folder_button.clicked.connect(self.scan_dicom_folder)

        import_layout.addWidget(scan_dicom_folder_button)
        import_layout.addWidget(self.importable_series_list)

        # Adding data_input columns to Main Layout
        layout.addWidget(scroll_area, stretch=1)
        layout.addWidget(import_group_box, stretch=1)
        self.data_tab.setLayout(layout)

    def dicom_import_to_folder(
        self,
        data_input: DataInputList,
        force_mod: bool = False,
        force_copy_list: list = None,
    ):
        """
        Copies the files inside the selected folder in the input list into the folder specified by data_input var.

        Parameters
        ----------
        data_input: DataInputList
            The name of the series to which couple the selected file.
        force_mod: bool
            Skip the modality check. Default is False
        force_copy_list: list
            If specificed, series is imported from here and not from scan list result for automatic import. Default is None

        Returns
        -------
        None.

        """

        if self.importable_series_list.currentRow() == -1 and force_copy_list is None:
            msg_box = QMessageBox()
            msg_box.setText(strings.subj_tab_selected_series_error)
            msg_box.exec()
            return

        origin = force_copy_list
        if origin is None:
            origin = self.dicom_scan_series_list[
                self.importable_series_list.currentRow()
            ]

        copy_list = origin[1]
        vols = origin[3]
        found_mod = origin[2].upper()

        progress = PersistentProgressDialog(
            strings.subj_tab_dicom_copy, 0, len(copy_list) + 1, self
        )
        self.set_loading(data_input)

        # Copy files and check for return
        import_ret = self.subject.dicom_import_to_folder(
            data_input=data_input,
            copy_list=copy_list,
            vols=vols,
            mod=found_mod,
            force_modality=force_mod,
            progress_callback=progress.increase_value,
        )
        if import_ret != SubjectRet.DataImportCompleted:
            if import_ret == SubjectRet.DataImportErrorVolumesMax:
                msg_box = QMessageBox()
                msg_box.setText(
                    strings.subj_tab_wrong_max_vols_check_msg
                    % (vols, data_input.value.max_volumes)
                )
                msg_box.exec()
            elif import_ret == SubjectRet.DataInputNonEmpty:
                msg_box = QMessageBox()
                msg_box.setText(strings.subj_tab_import_folder_not_empy)
                msg_box.exec()
            elif import_ret == SubjectRet.DataImportErrorVolumesMin:
                msg_box = QMessageBox()
                msg_box.setText(
                    strings.subj_tab_wrong_min_vols_check_msg
                    % (vols, data_input.value.min_volumes)
                )
                msg_box.exec()
            elif import_ret == SubjectRet.DataImportErrorCopy:
                msg_box = QMessageBox()
                msg_box.setText(strings.subj_tab_import_copy_error_msg)
                msg_box.exec()
            elif import_ret == SubjectRet.DataImportErrorModality:
                msg_box = QMessageBox()
                msg_box.setText(
                    strings.subj_tab_wrong_type_check_msg
                    % (found_mod, data_input.value.image_modality.value)
                )
                msg_box.setInformativeText(strings.subj_tab_wrong_type_check)
                msg_box.setIcon(QMessageBox.Icon.Warning)
                msg_box.setStandardButtons(
                    QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
                )
                msg_box.setDefaultButton(QMessageBox.StandardButton.No)
                ret = msg_box.exec()
                if ret == QMessageBox.StandardButton.Yes:
                    self.dicom_import_to_folder(data_input, force_mod=True)
            self.set_error(data_input, "")
            progress.deleteLater()
            return

        progress.setRange(0, 0)
        progress.setLabelText(strings.subj_tab_dicom_check)

        self.subject.check_input_folder(
            data_input,
            status_callback=self.input_check_update,
            progress_callback=progress.increase_value,
        )
        self.reset_workflow()

    def scan_dicom_folder(self):
        """
        Opens a folder dialog window to select the DICOM files folder to import.
        Scans the folder in a new thread.

        Returns
        -------
        None.

        """

        folder_path = QFileDialog.getExistingDirectory(
            self, strings.subj_tab_select_dicom_folder
        )

        if not os.path.exists(folder_path):
            return

        dicom_src_work = DicomSearchWorker(
            folder_path,
            classify=self.global_config.getboolean_safe(
                GlobalPrefCategoryList.MAIN, "auto_import"
            ),
        )
        dicom_src_work.load_dir()

        if dicom_src_work.get_files_len() > 0:
            self.clear_scan_result()
            self.dicom_scan_series_list = []
            progress = PersistentProgressDialog(
                strings.subj_tab_dicom_scan, 0, 0, parent=self.parent()
            )
            progress.show()
            progress.setMaximum(dicom_src_work.get_files_len() + 1)
            dicom_src_work.signal.sig_loop.connect(lambda i: progress.increase_value(i))
            dicom_src_work.signal.sig_finish.connect(self.show_scan_result)
            QThreadPool.globalInstance().start(dicom_src_work)

        else:
            msg_box = QMessageBox()
            msg_box.setText(strings.subj_tab_no_dicom_error + folder_path)
            msg_box.exec()

    def exec_tab_ui(self):
        """
        Generates the Execute Workflow tab UI.

        Returns
        -------
        None.

        """

        layout = QGridLayout()

        # First Column: NODE LIST
        self.workflow_type_combo = QComboBox(self)

        for row in WORKFLOW_TYPES:
            self.workflow_type_combo.addItem(row.value, userData=row.name)

        layout.addWidget(self.workflow_type_combo, 0, 0)

        self.generate_workflow_button = QPushButton(strings.GENBUTTONTEXT)
        self.generate_workflow_button.setFixedHeight(
            self.main_window.NON_UNICODE_BUTTON_HEIGHT
        )
        self.generate_workflow_button.setSizePolicy(
            QSizePolicy.Minimum, QSizePolicy.Fixed
        )
        self.generate_workflow_button.clicked.connect(self.generate_workflow)

        layout.addWidget(self.generate_workflow_button, 1, 0)

        self.node_list_treeWidget = QTreeWidget()
        self.node_list_treeWidget.setHeaderHidden(True)
        node_list_width = 320
        self.node_list_treeWidget.setFixedWidth(node_list_width)
        self.node_list_treeWidget.header().setMinimumSectionSize(node_list_width)
        self.node_list_treeWidget.header().setSectionResizeMode(
            QHeaderView.ResizeToContents
        )
        self.node_list_treeWidget.header().setStretchLastSection(False)
        self.node_list_treeWidget.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.node_list_treeWidget.horizontalScrollBar().setEnabled(True)

        layout.addWidget(self.node_list_treeWidget, 2, 0)
        self.node_list_treeWidget.itemClicked.connect(self.tree_item_clicked)

        # Second Column: Graphviz Graph Layout
        self.subject_config_button = QPushButton(strings.SUBJCONFIGBUTTONTEXT)
        self.subject_config_button.setFixedHeight(
            self.main_window.NON_UNICODE_BUTTON_HEIGHT
        )
        self.subject_config_button.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self.subject_config_button.clicked.connect(self.edit_subject_config)
        layout.addWidget(self.subject_config_button, 0, 1)

        self.exec_button = QPushButton(strings.EXECBUTTONTEXT)
        self.exec_button.setFixedHeight(self.main_window.NON_UNICODE_BUTTON_HEIGHT)
        self.exec_button.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self.exec_button.clicked.connect(
            partial(self.toggle_workflow_execution, None, None)
        )
        self.exec_button_set_enabled(False)

        layout.addWidget(self.exec_button, 1, 1)
        self.exec_graph = QSvgWidget()
        layout.addWidget(self.exec_graph, 2, 1)

        self.exec_tab.setLayout(layout)

    def edit_subject_config(self):
        """
        Opens the subject Preference Window.

        Returns
        -------
        None.

        """

        preference_window = PreferencesWindow(
            self.subject.config, self.subject.dependency_manager, True, self
        )
        ret = preference_window.exec()
        if ret != 0:
            self.reset_workflow()
        if ret == -1:
            self.subject.config.reset_to_defaults()
            self.edit_subject_config()

    def on_wf_type_changed(self, index: int):
        """
        Updates the workflow at workflow type combo change.

        Parameters
        ----------
        index : int
            The new selected value from the Execution tab workflow type combo.

        Returns
        -------
        None.

        """

        new_workflow_type = self.workflow_type_combo.itemData(index)
        self.subject.config.set_workflow_option(WORKFLOW_TYPES[new_workflow_type])
        self.subject.config.save()
        self.reset_workflow()

    def generate_workflow(self):
        """
        Generates and populates the Main Workflow.
        Shows the node list into the UI.
        Generates the graphviz analysis graphs on a new thread.

        Returns
        -------
        None.

        """

        generate_workflow_return = self.subject.generate_workflow()

        if generate_workflow_return == SubjectRet.GenWfMissingRequisites:
            error_dialog = QErrorMessage(parent=self)
            error_dialog.showMessage(strings.subj_tab_missing_fsl_error)
            return
        elif generate_workflow_return == SubjectRet.GenWfError:
            error_dialog = QErrorMessage(parent=self)
            error_dialog.showMessage(strings.subj_tab_wf_gen_error)
            return

        self.node_list_treeWidget.clear()
        self.node_list = self.subject.workflow.get_node_array()

        # Graphviz analysis graphs drawing
        for node in self.node_list.keys():
            self.node_list[node].node_holder = CustomTreeWidgetItem(
                self.node_list_treeWidget,
                self.node_list_treeWidget,
                self.node_list[node].long_name,
            )
            if len(self.node_list[node].node_list.keys()) > 0:
                for sub_node in self.node_list[node].node_list.keys():
                    self.node_list[node].node_list[sub_node].node_holder = (
                        CustomTreeWidgetItem(
                            self.node_list[node].node_holder,
                            self.node_list_treeWidget,
                            self.node_list[node].node_list[sub_node].long_name,
                        )
                    )

        # UI updating
        self.exec_button_set_enabled(True)
        self.generate_workflow_button.setEnabled(False)

    def tree_item_clicked(self, item, col: int):
        """
        Listener for the QTreeWidget Items.
        Shows the clicked analysis graphviz graph.

        Parameters
        ----------
        item : QTreeWidget Item
            The QTreeWidget item clicked.
        col : int
            The QTreeWidget column.

        Returns
        -------
        None.

        """

        if item.parent() is None:
            graph_file = self.subject.graph_file(item.get_text())
            if os.path.exists(graph_file):
                self.exec_graph.load(graph_file)
                self.exec_graph.renderer().setAspectRatioMode(
                    Qt.AspectRatioMode.KeepAspectRatio
                )

    @staticmethod
    def no_close_event(event):
        """
        Used to prevent the user to close a dialog.

        Parameters
        ----------
        event : TYPE
            The event to ignore.

        Returns
        -------
        None.

        """

        event.ignore()

    def toggle_workflow_execution(
        self, resume: bool = None, resume_freesurfer: bool = None
    ):
        """
        If the workflow is not started, executes it.
        If the workflow is executing, kills it.

        Returns
        -------
        None.

        """

        # Workflow not started
        if not self.subject.is_workflow_process_alive():
            workflow_start_ret = self.subject.start_workflow(
                resume=resume,
                resume_freesurfer=resume_freesurfer,
                update_node_callback=self.update_node_list,
            )
            if workflow_start_ret == SubjectRet.ExecWfResume:
                msg_box = QMessageBox()
                msg_box.setText(strings.subj_tab_old_wf_found)
                msg_box.setIcon(QMessageBox.Icon.Question)
                msg_box.setStandardButtons(
                    QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
                )
                msg_box.button(QMessageBox.StandardButton.Yes).setText(
                    strings.subj_tab_old_wf_resume
                )
                msg_box.button(QMessageBox.StandardButton.No).setText(
                    strings.subj_tab_old_wf_reset
                )
                msg_box.setDefaultButton(QMessageBox.StandardButton.Yes)
                msg_box.setWindowFlags(Qt.CustomizeWindowHint | Qt.WindowTitleHint)
                msg_box.closeEvent = self.no_close_event
                ret = msg_box.exec()
                resume = ret == QMessageBox.StandardButton.Yes
                self.toggle_workflow_execution(
                    resume=resume, resume_freesurfer=resume_freesurfer
                )
            elif workflow_start_ret == SubjectRet.ExecWfResumeFreesurfer:
                msg_box = QMessageBox()
                msg_box.setText(strings.subj_tab_old_fs_found)
                msg_box.setIcon(QMessageBox.Icon.Question)
                msg_box.setStandardButtons(
                    QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
                )
                msg_box.button(QMessageBox.StandardButton.Yes).setText(
                    strings.subj_tab_old_fs_resume
                )
                msg_box.button(QMessageBox.StandardButton.No).setText(
                    strings.subj_tab_old_fs_reset
                )
                msg_box.setDefaultButton(QMessageBox.StandardButton.Yes)
                msg_box.setWindowFlags(Qt.CustomizeWindowHint | Qt.WindowTitleHint)
                msg_box.closeEvent = self.no_close_event
                ret = msg_box.exec()
                resume_freesurfer = ret == QMessageBox.StandardButton.Yes
                self.toggle_workflow_execution(
                    resume=resume, resume_freesurfer=resume_freesurfer
                )
            elif workflow_start_ret == SubjectRet.ExecWfStatusError:
                # Already running, should not be possible
                pass
            else:
                # UI updating
                self.exec_button.setText(strings.EXECBUTTONTEXT_STOP)
                self.setTabEnabled(SubjectTab.DATATAB, False)
                self.setTabEnabled(SubjectTab.RESULTTAB, False)
                self.workflow_type_combo.setEnabled(False)
                self.subject_config_button.setEnabled(False)

        # Workflow executing
        else:
            # Asks for workflow kill confirmation
            msg_box = QMessageBox()
            msg_box.setText(strings.subj_tab_wf_stop)
            msg_box.setIcon(QMessageBox.Icon.Question)
            msg_box.setStandardButtons(
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No
            )
            msg_box.setDefaultButton(QMessageBox.StandardButton.No)
            msg_box.closeEvent = self.no_close_event
            ret = msg_box.exec()

            if ret == QMessageBox.StandardButton.No:
                return

            workflow_stop_ret = self.subject.stop_workflow()
            if workflow_stop_ret == SubjectRet.ExecWfStatusError:
                # Not running, should not be possible
                pass
            else:
                # UI updating
                self.remove_running_icon()
                self.exec_button.setText(strings.EXECBUTTONTEXT)
                self.setTabEnabled(SubjectTab.DATATAB, True)
                self.reset_workflow(force=True)
                self.enable_tab_if_result_dir()

    def export_results_button_update_state(self):
        """
        Enable the export results button if Slicer is found on system
        """
        try:
            if not DependencyManager.is_slicer(self.global_config):
                self.generate_scene_button.setEnabled(False)
                self.generate_scene_button.setToolTip(
                    strings.subj_tab_generate_scene_button_disabled_tooltip
                )
            else:
                self.generate_scene_button.setEnabled(True)
                self.generate_scene_button.setToolTip(
                    strings.subj_tab_generate_scene_button_tooltip
                )
        except:
            pass

    def load_scene_button_update_state(self):
        """
        Enable the load scene button if a scene file is present, otherwise disable it
        """
        try:
            if not DependencyManager.is_slicer(self.global_config):
                self.load_scene_button.setEnabled(False)
                self.load_scene_button.setText(
                    strings.subj_tab_load_scene_button + " " + strings.INFOCHAR
                )
                self.load_scene_button.setToolTip(
                    strings.subj_tab_generate_scene_button_disabled_tooltip
                )
            elif os.path.exists(self.subject.scene_path()):
                self.load_scene_button.setEnabled(True)
                self.load_scene_button.setToolTip("")
                self.load_scene_button.setText(strings.subj_tab_load_scene_button)
            else:
                self.load_scene_button.setEnabled(False)
                self.load_scene_button.setToolTip(
                    strings.subj_tab_load_scene_button_tooltip
                )
                self.load_scene_button.setText(
                    strings.subj_tab_load_scene_button + " " + strings.INFOCHAR
                )
        except:
            pass

    def result_tab_ui(self):
        """
        Generates the Results tab UI.

        Returns
        -------
        None.

        """

        result_tab_layout = QGridLayout()
        self.result_tab.setLayout(result_tab_layout)

        self.generate_scene_button = QPushButton(strings.subj_tab_generate_scene_button)
        self.generate_scene_button.clicked.connect(self.generate_scene)
        self.generate_scene_button.setFixedHeight(
            self.main_window.NON_UNICODE_BUTTON_HEIGHT
        )
        self.generate_scene_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.export_results_button_update_state()
        result_tab_layout.addWidget(self.generate_scene_button, 0, 0)

        horizontal_spacer = QSpacerItem(
            20, 40, QSizePolicy.Expanding, QSizePolicy.Minimum
        )
        result_tab_layout.addItem(horizontal_spacer, 0, 1, 1, 1)

        self.load_scene_button = QPushButton(strings.subj_tab_load_scene_button)
        self.load_scene_button.clicked.connect(self.load_scene)
        self.load_scene_button.setFixedHeight(
            self.main_window.NON_UNICODE_BUTTON_HEIGHT
        )
        self.load_scene_button.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        self.load_scene_button_update_state()
        result_tab_layout.addWidget(self.load_scene_button, 0, 2)

        self.open_results_directory_button = QPushButton(
            strings.subj_tab_open_results_directory
        )
        self.open_results_directory_button.clicked.connect(
            lambda pushed=False, results_dir=self.subject.result_dir(): open_results_directory(
                pushed, results_dir
            )
        )
        self.open_results_directory_button.setSizePolicy(
            QSizePolicy.Fixed, QSizePolicy.Fixed
        )
        self.open_results_directory_button.setFixedHeight(
            self.main_window.NON_UNICODE_BUTTON_HEIGHT
        )
        result_tab_layout.addWidget(self.open_results_directory_button, 0, 3)

        self.results_model = QFileSystemModel()
        self.result_tree = QTreeView(parent=self)
        self.result_tree.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.result_tree.setModel(self.results_model)

        result_tab_layout.addWidget(self.result_tree, 1, 0, 1, 4)

    def generate_scene(self) -> PersistentProgressDialog:
        """
        Exports the workflow results into 3D Slicer using a new thread.

        Returns
        -------
        The progress dialog shown.

        """

        progress = PersistentProgressDialog(
            strings.subj_tab_exporting_start, 0, 0, parent=self
        )
        progress.show()
        self.subject.generate_scene(
            lambda msg: SubjectTab.slicer_thread_signal(msg, progress)
        )
        return progress

    @staticmethod
    def slicer_thread_signal(msg: str, progress: PersistentProgressDialog):
        """
        Updates the Progress Dialog text to inform the user of the loading status.

        Parameters
        ----------
        msg : str
            The loading text.
        progress : PersistentProgressDialog
            The Progress Dialog.

        Returns
        -------
        None.

        """

        if msg == SlicerExportWorker.END_MSG:
            progress.done(1)
        else:
            progress.setLabelText(strings.subj_tab_exporting_prefix + msg)

    def input_check_update(
        self,
        data_input: DataInputList,
        state: SubjectRet,
        dicom_src_worker: DicomSearchWorker = None,
    ):
        """
        Function used as callback after subject dicom folder check

        Parameters
        ----------
        data_input: DataInputList
            The data input checked
        state: SubjectRet
            The return code of the scan
        dicom_src_worker: DicomSearchWorker, optional
            The dicom scanner calling the function. Default is None
        """
        if data_input not in self.input_report:
            return
        if state == SubjectRet.DataInputWarningNoDicom:
            self.set_error(
                data_input, strings.subj_tab_no_dicom_error + dicom_src_worker.dicom_dir
            )
        elif state == SubjectRet.DataInputWarningMultiSubj:
            self.set_warn(
                data_input,
                strings.subj_tab_multi_subj_error + dicom_src_worker.dicom_dir,
            )
        elif state == SubjectRet.DataInputWarningMultiStudy:
            self.set_warn(
                data_input,
                strings.subj_tab_multi_exam_error + dicom_src_worker.dicom_dir,
            )
        elif state == SubjectRet.DataInputWarningMultiSeries:
            self.set_warn(
                data_input,
                strings.subj_tab_multi_series_error + dicom_src_worker.dicom_dir,
            )
        elif state == SubjectRet.DataInputLoading:
            self.set_loading(data_input)
        elif state == SubjectRet.DataInputValid:

            subject_list = dicom_src_worker.tree.get_subject_list()
            exam_list = dicom_src_worker.tree.get_studies_list(subject_list[0])
            series_list = dicom_src_worker.tree.get_series_list(
                subject_list[0], exam_list[0]
            )
            series = dicom_src_worker.tree.get_series(
                subject_list[0], exam_list[0], series_list[0]
            )
            subject_name = dicom_src_worker.tree.dicom_subjects[
                subject_list[0]
            ].subject_name
            series_description = series.description
            vols = series.volumes
            mod = series.modality
            frames = series.frames

            label = SubjectTab.label_from_dicom(
                frames, subject_name, mod, series_description, vols
            )

            self.set_ok(data_input, label)
            self.enable_exec_tab()
            self.check_venous_volumes()

    def load_subject(self, check_dicom_folders: bool = True):
        """
        Loads the subject configuration and folder.

        Parameters
        ----------
        check_dicom_folders : bool
            If True, check for dicom files in the fubject folders. Default is True.

        Returns
        -------
        None.

        """

        index = self.workflow_type_combo.findData(
            self.subject.config.get_subject_workflow_type().name
        )
        self.workflow_type_combo.setCurrentIndex(index)
        # Set after subject loading to prevent the onchanged fire on previous line command
        self.workflow_type_combo.currentIndexChanged.connect(self.on_wf_type_changed)

        if check_dicom_folders:
            # Scan subject dicom folder
            dicom_scanners, total_files = self.subject.prepare_scan_dicom_folders()

            if total_files > 0:
                progress = PersistentProgressDialog(
                    strings.subj_tab_subj_loading, 0, 0, parent=self.parent()
                )
                progress.show()
                progress.setMaximum(total_files)
                self.subject.execute_scan_dicom_folders(
                    dicom_scanners,
                    status_callback=self.input_check_update,
                    progress_callback=progress.increase_value,
                )

        # Update UI after loading dicom
        self.setTabEnabled(SubjectTab.DATATAB, True)
        self.setCurrentWidget(self.data_tab)

        self.clear_scan_result()
        self.reset_workflow()

        self.enable_tab_if_result_dir()

    def enable_tab_if_result_dir(
        self, changed_path: str = "", on_workflow_running: bool = False
    ):
        """
        Enables Results tab, if any.

        Parameters
        _______
        changed_path: str
            The changed directory, passed if called by watcher
        on_workflow_running: bool
            If True, updte states even if workflow is running. Default is False.

        Returns
        -------
        None.

        """
        scene_dir = self.subject.result_dir()

        if self.subject.is_workflow_process_alive() and not on_workflow_running:
            return

        if os.path.exists(scene_dir):
            self.setTabEnabled(SubjectTab.RESULTTAB, True)
            self.results_model.setRootPath(scene_dir)
            index_root = self.results_model.index(self.results_model.rootPath())
            self.result_tree.setRootIndex(index_root)
            self.result_directory_watcher.addPath(scene_dir)
            self.load_scene_button_update_state()
        else:
            self.setTabEnabled(SubjectTab.RESULTTAB, False)
            self.result_directory_watcher.removePaths([scene_dir])

    def clear_import_folder(self, data_input: DataInputList):
        """
        Clears the subject series folder.

        Parameters
        ----------
        data_input : DataInputList
            The series folder name to clear.

        Returns
        -------
        None.

        """

        src_path = self.subject.dicom_folder(data_input)

        progress = PersistentProgressDialog(
            strings.subj_tab_dicom_clearing + src_path, 0, 0, self
        )
        progress.show()

        self.subject.clear_import_folder(data_input)

        self.set_error(data_input, strings.subj_tab_no_dicom_error + src_path)
        self.enable_exec_tab()

        progress.accept()

        self.reset_workflow()
        self.check_venous_volumes()

        if (
            data_input == DataInputList.VENOUS
            and self.subject.input_state_list[DataInputList.VENOUS2].loaded
        ):
            self.clear_import_folder(DataInputList.VENOUS2)

    def check_venous_volumes(self):
        """
        Display informative warnings in venous and venous2 data input rows to help user in data loading
        """
        phases = (
            self.subject.input_state_list[DataInputList.VENOUS].volumes
            + self.subject.input_state_list[DataInputList.VENOUS2].volumes
        )
        if phases == 0:
            self.input_report[DataInputList.VENOUS2][3].setEnabled(False)
        elif phases == 1:
            if self.subject.input_state_list[DataInputList.VENOUS].loaded:
                self.set_warn(
                    DataInputList.VENOUS,
                    "Series has only one phase, load the second phase below",
                    False,
                )
                self.input_report[DataInputList.VENOUS2][3].setEnabled(True)
            if self.subject.input_state_list[DataInputList.VENOUS2].loaded:
                # this should not be possible!
                self.set_warn(
                    DataInputList.VENOUS2,
                    "Series has only one phase, load the second phase above",
                    False,
                )
        elif phases == 2:
            if self.subject.input_state_list[DataInputList.VENOUS].loaded:
                self.set_ok(DataInputList.VENOUS, None)
                self.input_report[DataInputList.VENOUS2][3].setEnabled(False)
            if self.subject.input_state_list[DataInputList.VENOUS2].loaded:
                self.set_ok(DataInputList.VENOUS2, None)
        else:
            # something gone wrong, more than 2 phases!
            if self.subject.input_state_list[DataInputList.VENOUS].loaded:
                self.set_warn(
                    DataInputList.VENOUS,
                    "Too many venous phases loaded, delete some!",
                    False,
                )
                self.input_report[DataInputList.VENOUS2][3].setEnabled(True)
            if self.subject.input_state_list[DataInputList.VENOUS2].loaded:
                self.set_warn(
                    DataInputList.VENOUS2,
                    "Too many venous phases loaded, delete some!",
                    False,
                )

    def exec_button_set_enabled(self, enabled: bool):
        """
        Change the status and the tooltip of the exec workflow button.

        Parameters
        ----------
        enabled : bool
            The new status of the button
        """
        if enabled:
            self.exec_button.setEnabled(True)
            self.exec_button.setToolTip("")
            self.exec_button.setText(strings.EXECBUTTONTEXT)
        else:
            self.exec_button.setEnabled(False)
            self.exec_button.setToolTip(strings.EXECBUTTONTEXT_disabled_tooltip)
            self.exec_button.setText(strings.EXECBUTTONTEXT + "  " + strings.INFOCHAR)

    def reset_workflow(self, force: bool = False):
        """
        Set the workflow var to None.
        Resets the UI.
        Works only if the worklow is not in execution or if force var is True.

        Parameters
        ----------
        force : bool, optional
            Force the usage of this function during workflow execution. The default is False.

        Returns
        -------
        None.

        """

        if self.subject.reset_workflow(force):
            self.node_list_treeWidget.clear()
            self.exec_graph.load(self.main_window.VOID_SVG_FILE)
            self.exec_button_set_enabled(False)
            self.generate_workflow_button.setEnabled(True)
            self.workflow_type_combo.setEnabled(True)
            self.subject_config_button.setEnabled(True)
            self.workflow_had_error = False

    @staticmethod
    def label_from_dicom(
        frames: int, subject_name: str, mod: str, series_description: str, vols: int
    ) -> str:
        """
        Compose dicom scan result into a readable label

        Parameters
        ----------
        image_list : list[str]
            The list of file found
        subject_name: str
            The subject name found
        mod: str
            The image modality found
        series_description: str
            The series description found
        vols: int
            The volumes count found in the series

        Returns
        -------
        A string containing the label text

        """
        try:
            # TODO: show correct image number for multiframe dicom
            label = (
                subject_name
                + "-"
                + mod
                + "-"
                + series_description
                + ": "
                + str(frames)
                + " images, "
                + str(vols)
                + " "
            )
            if vols > 1:
                label += "volumes"
            else:
                label += "volume"
            return label
        except:
            return ""

    def show_scan_result(self, dicom_src_work: DicomSearchWorker):
        """
        Updates importable series list using DICOM Search Worker results.

        Parameters
        ----------
        dicom_src_work : DicomSearchWorker
            The DICOM Search Worker.

        Returns
        -------
        None.

        """

        folder_path = dicom_src_work.dicom_dir
        self.scan_directory_watcher.addPath(folder_path)
        subject_list = dicom_src_work.tree.get_subject_list()

        if len(subject_list) == 0:
            msg_box = QMessageBox()
            if len(dicom_src_work.error_message) > 0:
                msg_box.setText(
                    strings.subj_tab_unsupported_files.format(
                        str(dicom_src_work.error_message)
                    )
                )
            else:
                msg_box.setText(strings.subj_tab_no_dicom_error + folder_path)
            msg_box.exec()
            return

        if len(subject_list) > 1:
            msg_box = QMessageBox()
            msg_box.setText(strings.subj_tab_multi_subj_error + folder_path)
            msg_box.exec()
            return

        studies_list = dicom_src_work.tree.get_studies_list(subject_list[0])

        for study in studies_list:
            series_list = dicom_src_work.tree.get_series_list(subject_list[0], study)
            for series in series_list:
                dicom_series = dicom_src_work.tree.get_series(
                    subject_list[0], study, series
                )
                frames = dicom_series.frames
                if frames == 0:
                    continue
                subject_name = dicom_src_work.tree.dicom_subjects[
                    subject_list[0]
                ].subject_name
                mod = dicom_series.modality
                series_description = dicom_series.description
                vols = dicom_series.volumes
                label = SubjectTab.label_from_dicom(
                    frames, subject_name, mod, series_description, vols
                )

                self.dicom_scan_series_list.append(
                    [
                        label,
                        dicom_series.dicom_locs,
                        mod,
                        vols,
                        dicom_series.classification,
                    ]
                )

        for series in self.dicom_scan_series_list:
            self.importable_series_list.addItem(series[0])

        if self.global_config.getboolean_safe(
            GlobalPrefCategoryList.MAIN, "auto_import"
        ):
            for data_input in self.subject.input_state_list:
                if not self.subject.input_state_list[data_input].loaded:
                    for series in self.dicom_scan_series_list:
                        if series[4] == data_input.value.name:
                            msg_box = QMessageBox()
                            msg_box.setText(
                                strings.subj_tab_found_series_type.format(
                                    series_description=series[0],
                                    data_label=data_input.value.label,
                                )
                            )
                            msg_box.setIcon(QMessageBox.Icon.Question)
                            msg_box.setStandardButtons(
                                QMessageBox.StandardButton.Yes
                                | QMessageBox.StandardButton.No
                            )
                            msg_box.setDefaultButton(QMessageBox.StandardButton.Yes)
                            msg_box.setWindowFlags(
                                Qt.CustomizeWindowHint | Qt.WindowTitleHint
                            )
                            msg_box.closeEvent = self.no_close_event
                            ret = msg_box.exec()
                            import_ret = ret == QMessageBox.StandardButton.Yes
                            if import_ret:
                                self.dicom_import_to_folder(
                                    data_input, force_copy_list=series
                                )
                                break

    def clear_scan_result(self):
        """
        Clear the content of the scan result list
        """
        self.importable_series_list.clear()
        self.dicom_scan_series_list = None
        if len(self.scan_directory_watcher.directories()) > 0:
            self.scan_directory_watcher.removePaths(
                self.scan_directory_watcher.directories()
            )

    def is_data_loading(self) -> bool:
        """
        Returns
            True if any data folder is being scanned
        """

        for row in self.input_report.values():
            if row[5] == self.main_window.LOADING_MOVIE_FILE:
                return True
        return False

    def update_input_report(
        self,
        data_input: DataInputList,
        icon: str,
        tooltip: str,
        import_enable: bool,
        clear_enable: bool,
        text: str = None,
    ):
        """
        Generic update function for series labels.

        Parameters
        ----------
        data_input: DataInputList
            The series label.
        icon: str
            The icon file to set near the label
        tooltip: str
            Mouse over tooltip:
        import_enable: bool
            The enable status of the import series button
        clear_enable: bool
            The enable status of the clear series button
        text: str
            The text to show under the label, if not None. Default is None
        """
        self.input_report[data_input][0].load(icon)
        self.input_report[data_input][0].setFixedSize(25, 25)
        self.input_report[data_input][0].setToolTip(tooltip)
        self.input_report[data_input][3].setEnabled(import_enable)
        self.input_report[data_input][4].setEnabled(clear_enable)
        self.input_report[data_input][5] = icon
        if text is not None:
            self.input_report[data_input][2].setText(text)

    def set_warn(
        self, data_input: DataInputList, tooltip: str, clear_text: bool = True
    ):
        """
        Set a warning message and icon near a series label.

        Parameters
        ----------
        data_input : DataInputList
            The series label.
        tooltip : str
            The warning message.
        clear_text : bool
            If True delete the label text

        Returns
        -------
        None.

        """
        text = None
        if clear_text:
            text = ""

        self.update_input_report(
            data_input=data_input,
            icon=self.main_window.WARNING_ICON_FILE,
            tooltip=tooltip,
            import_enable=False,
            clear_enable=True,
            text=text,
        )

    def set_error(self, data_input: DataInputList, tooltip: str):
        """
        Set an error message and icon near a series label.

        Parameters
        ----------
        data_input : DataInputList
            The series label.
        tooltip : str
            The error message.
        """

        self.update_input_report(
            data_input=data_input,
            icon=self.main_window.ERROR_ICON_FILE,
            tooltip=tooltip,
            import_enable=True,
            clear_enable=False,
            text="",
        )

    def set_ok(self, data_input: DataInputList, text: str):
        """
        Set a success message and icon near a series label.

        Parameters
        ----------
        data_input : DataInputList
            The series label.
        text : str
            The success message. If string is None keep the current text
        """
        self.update_input_report(
            data_input=data_input,
            icon=self.main_window.OK_ICON_FILE,
            tooltip="",
            import_enable=False,
            clear_enable=True,
            text=text,
        )

    def set_loading(self, data_input: DataInputList):
        """
        Set a loading message and icon near a series label.

        Parameters
        ----------
        data_input : DataInputList
            The series label.
        """
        self.update_input_report(
            data_input=data_input,
            icon=self.main_window.LOADING_MOVIE_FILE,
            tooltip="",
            import_enable=False,
            clear_enable=False,
            text=None,
        )

    def enable_exec_tab(self):
        """
        Enables the Execute Workflow tab into the UI.

        Returns
        -------
        None.

        """

        enable = self.subject.can_generate_workflow()
        self.setTabEnabled(SubjectTab.EXECTAB, enable)

    def load_scene(self):
        """
        Visualize the workflow results into 3D Slicer.
        """

        slicer_open_thread = SlicerViewerWorker(
            self.global_config.get_slicer_path(), self.subject.scene_path()
        )
        QThreadPool.globalInstance().start(slicer_open_thread)

    def setTabEnabled(self, index: int, enabled: bool):
        """
        Changes the status of a tab and set an informative tooltip

        Parameters
        -------
        index: int
            The tab index
        enabled: bool
            The new tab status
        """
        if index == SubjectTab.EXECTAB and not enabled:
            if (
                not self.subject.dependency_manager.is_fsl()
                or not self.subject.dependency_manager.is_dcm2niix()
            ):
                self.setTabToolTip(
                    index, strings.subj_tab_tabtooltip_exec_disabled_dependency
                )
            else:
                self.setTabToolTip(
                    index, strings.subj_tab_tabtooltip_exec_disabled_series
                )
        elif index == SubjectTab.RESULTTAB and not enabled:
            self.setTabToolTip(index, strings.subj_tab_tabtooltip_result_disabled)
        elif index == SubjectTab.DATATAB and not enabled:
            self.setTabToolTip(index, strings.subj_tab_tabtooltip_data_disabled)
        else:
            self.setTabToolTip(index, "")
        super().setTabEnabled(index, enabled)

    def setTabToolTip(self, index: int, tooltip: str):
        """
        Changes the tooltip of a tab

        Parameters
        -------
        index: int
            The tab index
        tooltip: str
            The tooltip to show
        """
        super().setTabToolTip(index, tooltip)
        if tooltip == "" and self.tabText(index).endswith(strings.INFOCHAR):
            self.setTabText(
                index, self.tabText(index).replace(" " + strings.INFOCHAR, "")
            )
        elif tooltip != "" and not self.tabText(index).endswith(strings.INFOCHAR):
            self.setTabText(index, self.tabText(index) + " " + strings.INFOCHAR)



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from nipype.interfaces.base import (
    traits,
    BaseInterface,
    BaseInterfaceInputSpec,
    TraitedSpec,
    File,
)


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterfaceInputSpec)  -*-
class RandomSeedGeneratorInputSpec(BaseInterfaceInputSpec):
    seeds_n = traits.Int(mandatory=True, desc="The number of needed seeds")
    mask = File(mandatory=True, exists=True, desc="Just for depend")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class RandomSeedGeneratorOutputSpec(TraitedSpec):
    seeds = traits.List(desc="the list of seeds")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterface)  -*-
class RandomSeedGenerator(BaseInterface):
    """
    Generates a list of random integers. Used to obtains a series of random seeds.

    """

    input_spec = RandomSeedGeneratorInputSpec
    output_spec = RandomSeedGeneratorOutputSpec
    seed_list = []

    def _run_interface(self, runtime):
        from random import randrange

        for x in range(self.inputs.seeds_n):
            self.seed_list.append(randrange(1000))

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["seeds"] = self.seed_list
        return outputs



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from os.path import abspath
import os
import glob
from nipype.interfaces.base import (
    traits,
    TraitedSpec,
    CommandLineInputSpec,
    CommandLine,
    File,
    Directory,
    isdefined,
)


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.CommandLineInputSpec)  -*-
class SegmentHAInputSpec(CommandLineInputSpec):
    subject_id = traits.Str(
        "recon_all",
        mandatory=True,
        position=0,
        argstr="%s",
        desc="subject name",
        usedefault=True,
    )
    subjects_dir = Directory(
        exists=True,
        mandatory=True,
        position=1,
        argstr="%s",
        hash_files=False,
        desc="path to subjects directory",
        genfile=True,
    )
    num_cpu = traits.Int(argstr="")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class SegmentHAOutputSpec(TraitedSpec):
    lh_hippoAmygLabels = File(
        desc="Discrete segmentation volumes at subvoxel resolution"
    )
    rh_hippoAmygLabels = File(
        desc="Discrete segmentation volumes at subvoxel resolution"
    )


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.CommandLine)  -*-
class SegmentHA(CommandLine):
    """
    Executes the segmentHA_T1.sh FreeSurfer command.

    """

    _cmd = "segmentHA_T1.sh"
    input_spec = SegmentHAInputSpec
    output_spec = SegmentHAOutputSpec

    def _list_outputs(self):
        base = os.path.join(self.inputs.subjects_dir, self.inputs.subject_id, "mri")
        lh = ""
        rh = ""

        src = glob.glob(
            os.path.abspath(os.path.join(base, "lh.hippoAmygLabels-T1.v[0-9][0-9].mgz"))
        )
        if len(src) == 1:
            lh = src[0]

        src = glob.glob(
            os.path.abspath(os.path.join(base, "rh.hippoAmygLabels-T1.v[0-9][0-9].mgz"))
        )
        if len(src) == 1:
            rh = src[0]

        return {"lh_hippoAmygLabels": lh, "rh_hippoAmygLabels": rh}

    def _parse_inputs(self, skip=None):
        """
        Custom implementation of _parse_inputs func to manage multithreading.

        """

        if isdefined(self.inputs.num_cpu):
            skip = ["num_cpu"]
            # self.n_procs = self.inputs.num_threads
            self.inputs.environ["ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS"] = (
                "%d" % self.inputs.num_cpu
            )

        parse = super(SegmentHA, self)._parse_inputs(skip)

        # Delete lock file if exists from previous execution
        ex_path = abspath(
            os.path.join(
                self.inputs.subjects_dir,
                self.inputs.subject_id,
                "scripts/IsRunningHPsubT1.lh+rh",
            )
        )

        if os.path.exists(ex_path):
            os.remove(ex_path)

        return parse



from PySide6.QtCore import QRunnable, Signal, QObject
import os
import subprocess

from swane.config.ConfigManager import ConfigManager
from swane.utils.DataInputList import DataInputList


class SlicerExportSignaler(QObject):
    export = Signal(str)


class SlicerExportWorker(QRunnable):
    """
    Spawn a thread for 3D Slicer result export

    """

    PROGRESS_MSG_PREFIX = "SLICERLOADER: "
    END_MSG = "ENDLOADING"

    def __init__(
        self, slicer_path: str, result_dir: str, scene_ext: str, config: ConfigManager
    ):
        super(SlicerExportWorker, self).__init__()
        self.signal = SlicerExportSignaler()
        self.slicer_path: str = slicer_path
        self.result_dir: str = result_dir
        self.scene_ext: str = scene_ext
        self.config: ConfigManager = config

    def run(self):

        vein_threshold = self.config.getfloat_safe(
            DataInputList.VENOUS, "vein_segment_threshold"
        )
        dti_threshold = self.config.getfloat_safe(
            DataInputList.DTI, "tractography_threshold"
        )

        cmd = (
            self.slicer_path
            + " --no-splash --no-main-window --python-script "
            + os.path.join(
                os.path.dirname(__file__),
                f"slicer_script_result.py {self.scene_ext} {str(dti_threshold)} {str(vein_threshold)}",
            )
        )

        popen = subprocess.Popen(
            cmd,
            cwd=self.result_dir,
            shell=True,
            stdout=subprocess.PIPE,
            universal_newlines=True,
        )
        for stdout_line in iter(popen.stdout.readline, ""):
            if stdout_line.startswith(self.PROGRESS_MSG_PREFIX):
                self.signal.export.emit(
                    stdout_line.replace(self.PROGRESS_MSG_PREFIX, "").replace("\n", "")
                )
        popen.stdout.close()
        popen.wait()
        self.signal.export.emit(self.END_MSG)



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from nipype.interfaces.fsl import BinaryMaths

from os.path import abspath
import os

from nipype.interfaces.base import (
    BaseInterface,
    BaseInterfaceInputSpec,
    TraitedSpec,
    File,
    isdefined,
)


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterfaceInputSpec)  -*-
class AsymmetryIndexInputSpec(BaseInterfaceInputSpec):

    in_file = File(exists=True, mandatory=True, desc="the input image")
    swapped_file = File(exists=True, mandatory=True, desc="the swapped input image")
    out_file = File(desc="the output image")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class AsymmetryIndexOutputSpec(TraitedSpec):
    out_file = File(desc="the output image")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterface)  -*-
class AsymmetryIndex(BaseInterface):
    """
    Generate Asymmetry Index Map from an image and its RL swapped as subtraction/sum.

    """

    input_spec = AsymmetryIndexInputSpec
    output_spec = AsymmetryIndexOutputSpec

    def _run_interface(self, runtime):
        self.inputs.out_file = self._gen_outfilename()

        add = BinaryMaths()
        add.inputs.in_file = self.inputs.in_file
        add.inputs.operand_file = self.inputs.swapped_file
        add.inputs.operation = "add"
        add.inputs.out_file = abspath("add_" + os.path.basename(self.inputs.in_file))
        add_res = add.run()

        sub = BinaryMaths()
        sub.inputs.in_file = self.inputs.in_file
        sub.inputs.operand_file = self.inputs.swapped_file
        sub.inputs.operation = "sub"
        sub.inputs.out_file = abspath("sub_" + os.path.basename(self.inputs.in_file))
        sub_res = sub.run()

        div = BinaryMaths()
        div.inputs.in_file = sub_res.outputs.out_file
        div.inputs.operand_file = add_res.outputs.out_file
        div.inputs.operation = "div"
        div.inputs.out_file = self.inputs.out_file
        div.run()

        return runtime

    def _gen_outfilename(self):
        out_file = self.inputs.out_file
        if not isdefined(out_file) and isdefined(self.inputs.in_file):
            out_file = "Aindex_" + os.path.basename(self.inputs.in_file)
        return abspath(out_file)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self._gen_outfilename()
        return outputs



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from nipype.interfaces.dcm2nii import Dcm2niix, Dcm2niixInputSpec
import os
from nipype.pipeline.engine.nodes import NodeExecutionError
from nipype.interfaces.base import traits


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.dcm2nii.Dcm2niixInputSpec)  -*-
class CustomDcm2niixInputSpec(Dcm2niixInputSpec):
    name_conflicts = traits.Enum(
        2,
        1,
        0,
        argstr="-w %d",
        usedefault=True,
        descr="write behavior for name conflicts - [0=skip duplicates, 1=overwrite, 2=add suffix]",
    )
    expected_files = traits.Int(default_value=1, usedefault=True)
    request_dti = traits.Bool(default_value=False, usedefault=True)


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.dcm2nii.Dcm2niix)  -*-
class CustomDcm2niix(Dcm2niix):
    """
    Custom implementation of Dcm2niix Nipype Node to support crop and merge parameters.

    """

    input_spec = CustomDcm2niixInputSpec

    def _run_interface(self, runtime):
        runtime = super(CustomDcm2niix, self)._run_interface(runtime)

        # Expected files check
        if (
            self.inputs.expected_files > 0
            and len(self.output_files) != self.inputs.expected_files
        ):
            raise NodeExecutionError(
                "Dcm2niix generated %d nifti files while %s were expected"
                % (len(self.output_files), self.inputs.expected_files)
            )

        # Bvec and Bvals check
        if self.inputs.request_dti and (len(self.bvals) == 0 or len(self.bvecs) == 0):
            raise NodeExecutionError(
                "Dcm2niix could not generate requested bvals and bvecs files"
            )

        return runtime



from nipype import Node, IdentityInterface, Merge, SelectFiles
from nipype.algorithms.modelgen import SpecifyModel
from nipype.algorithms.rapidart import ArtifactDetect
from nipype.interfaces.fsl import (
    ImageMaths,
    ExtractROI,
    MCFLIRT,
    BET,
    ImageStats,
    SUSAN,
    FLIRT,
    Level1Design,
    FEATModel,
    FILMGLS,
    SmoothEstimate,
    Cluster,
    ApplyXFM,
)
from configparser import SectionProxy
from swane.nipype_pipeline.engine.CustomWorkflow import CustomWorkflow
from swane.nipype_pipeline.nodes.CustomDcm2niix import CustomDcm2niix
from swane.nipype_pipeline.nodes.FslNVols import FslNVols
from swane.nipype_pipeline.nodes.FMRIGenSpec import FMRIGenSpec
from swane.nipype_pipeline.nodes.CustomSliceTimer import CustomSliceTimer
from swane.nipype_pipeline.nodes.GetNiftiTR import GetNiftiTR
from swane.nipype_pipeline.nodes.ForceOrient import ForceOrient
from swane.nipype_pipeline.nodes.DeleteVolumes import DeleteVolumes
from swane.config.config_enums import BLOCK_DESIGN


def task_fMRI_workflow(
    name: str, dicom_dir: str, config: SectionProxy, base_dir: str = "/"
) -> CustomWorkflow:
    """
    fMRI first level anlysis for a single task with constant task-rest paradigm.

    Parameters
    ----------
    name : str
        The workflow name.
    dicom_dir : path
        The directory path of the DICOM files.
    config: SectionProxy
        workflow settings.
    base_dir : path, optional
        The base directory path relative to parent workflow. The default is "/".

    Input Node Fields
    ----------
    ref_BET : path
        Betted T13D.

    Output Node Fields
    ----------
    threshold_file_1 : path
        Cluster of activation (task A vs rest or Task A vs Task B) in T13D reference space.
    threshold_file_2 : path
        Cluster of activation (task b vs Task) in T13D reference space.

    Returns
    -------
    workflow : CustomWorkflow
        The fMRI workflow.

    """

    workflow = CustomWorkflow(name=name, base_dir=base_dir)

    # Input Node
    inputnode = Node(IdentityInterface(fields=["ref_BET"]), name="inputnode")

    task_a_name = config["task_a_name"].replace(" ", "_")
    task_b_name = config["task_b_name"].replace(" ", "_")
    task_duration = config.getint_safe("task_duration")
    rest_duration = config.getint_safe("rest_duration")
    TR = config.getfloat_safe("tr")
    slice_timing = config.getenum_safe("slice_timing")
    n_vols = config.getint_safe("n_vols")
    del_start_vols = config.getint_safe("del_start_vols")
    del_end_vols = config.getint_safe("del_end_vols")
    block_design = config.getenum_safe("block_design")

    # Output Node
    outputnode = Node(
        IdentityInterface(
            fields=[
                "threshold_file_cont1_thresh1",
                "threshold_file_cont1_thresh2",
                "threshold_file_cont1_thresh3",
                "threshold_file_cont2_thresh1",
                "threshold_file_cont2_thresh2",
                "threshold_file_cont2_thresh3",
            ]
        ),
        name="outputnode",
    )

    # NODE 1: Conversion dicom -> nifti
    conversion = Node(CustomDcm2niix(), name="%s_conv" % name)
    conversion.inputs.out_filename = name
    conversion.inputs.bids_format = False
    conversion.inputs.source_dir = dicom_dir
    conversion.inputs.name_conflicts = 1
    conversion.inputs.merge_imgs = 2

    # NODE 2: Get EPI volume numbers
    nvols = Node(FslNVols(), name="%s_nvols" % name)
    nvols.long_name = "EPI volumes count"
    nvols.inputs.force_value = n_vols
    workflow.connect(conversion, "converted_files", nvols, "in_file")

    # NODE 3: Get Repetition Time
    getTR = Node(GetNiftiTR(), name="%s_getTR" % name)
    getTR.long_name = "get TR"
    getTR.inputs.force_value = TR
    workflow.connect(conversion, "converted_files", getTR, "in_file")

    # NODE 4: Delete specified volumes at start and end of sequence
    del_vols = Node(DeleteVolumes(), name="%s_del_vols" % name)
    del_vols.long_name = "Edge volumes trimming"
    del_vols.inputs.del_start_vols = del_start_vols
    del_vols.inputs.del_end_vols = del_end_vols
    workflow.connect(conversion, "converted_files", del_vols, "in_file")
    workflow.connect(nvols, "nvols", del_vols, "nvols")

    # NODE 5: Orienting in radiological convention
    reorient = Node(ForceOrient(), name="%s_reorient" % name)
    workflow.connect(del_vols, "out_file", reorient, "in_file")

    # NODE 5: Convert functional images to float representation.
    img2float = Node(ImageMaths(), name="%s_img2float" % name)
    img2float.long_name = "Intensity in float values"
    img2float.inputs.out_data_type = "float"
    img2float.inputs.op_string = ""
    img2float.inputs.suffix = "_dtype"
    workflow.connect(reorient, "out_file", img2float, "in_file")

    # NODE 6: Extract the middle volume of the first run as the reference
    extract_ref = Node(ExtractROI(), name="%s_extract_ref" % name)
    extract_ref.long_name = "Reference volume selection"
    extract_ref.inputs.t_size = 1

    # Function to extract the middle volume number
    def get_middle_volume(func):
        from nibabel import load

        funcfile = func
        if isinstance(func, list):
            funcfile = func[0]
        _, _, _, timepoints = load(funcfile).shape
        middle = int(timepoints / 2)
        return middle

    workflow.connect(img2float, "out_file", extract_ref, "in_file")
    workflow.connect(reorient, ("out_file", get_middle_volume), extract_ref, "t_min")

    # NODE 7: Realign the functional runs to the middle volume of the first run
    motion_correct = Node(MCFLIRT(), name="%s_motion_correct" % name)
    motion_correct.inputs.save_mats = True
    motion_correct.inputs.save_plots = True
    motion_correct.inputs.save_rms = True
    motion_correct.inputs.interpolation = "spline"
    workflow.connect(img2float, "out_file", motion_correct, "in_file")
    workflow.connect(extract_ref, "roi_file", motion_correct, "ref_file")

    # NODE 8: Perform slice timing correction if needed
    slice_timing_correction = Node(
        CustomSliceTimer(), name="%s_timing_correction" % name
    )
    slice_timing_correction.inputs.slice_timing = slice_timing
    workflow.connect(getTR, "TR", slice_timing_correction, "time_repetition")
    workflow.connect(motion_correct, "out_file", slice_timing_correction, "in_file")

    # NODE 9: Extract the mean volume of the first functional run
    meanfunc = Node(ImageMaths(), name="%s_meanfunc" % name)
    meanfunc.long_name = "mean image calculation"
    meanfunc.inputs.op_string = "-Tmean"
    meanfunc.inputs.suffix = "_mean"
    workflow.connect(
        slice_timing_correction, "slice_time_corrected_file", meanfunc, "in_file"
    )

    # NODE 10: Strip the skull from the mean functional to generate a mask
    meanfuncmask = Node(BET(), name="%s_meanfuncmask" % name)
    meanfuncmask.inputs.mask = True
    meanfuncmask.inputs.no_output = True
    meanfuncmask.inputs.frac = 0.3
    workflow.connect(meanfunc, "out_file", meanfuncmask, "in_file")

    # NODE 11: Mask the functional runs with the extracted mask
    maskfunc = Node(ImageMaths(), name="%s_maskfunc" % name)
    maskfunc.long_name = "mean image masking"
    maskfunc.inputs.suffix = "_bet"
    maskfunc.inputs.op_string = "-mas"
    workflow.connect(
        slice_timing_correction, "slice_time_corrected_file", maskfunc, "in_file"
    )
    workflow.connect(meanfuncmask, "mask_file", maskfunc, "in_file2")

    # NODE 12: Determine the 2nd and 98th percentile intensities of each functional run
    getthresh = Node(ImageStats(), name="%s_getthresh" % name)
    getthresh.long_name = "2-98% threshold calculation"
    getthresh.inputs.op_string = "-p 2 -p 98"
    workflow.connect(maskfunc, "out_file", getthresh, "in_file")

    # NODE 13: Threshold the first run of the functional data at 10% of the 98th percentile
    threshold = Node(ImageMaths(), name="%s_threshold" % name)
    threshold.long_name = "thresholding"
    threshold.inputs.out_data_type = "char"
    threshold.inputs.suffix = "_thresh"

    # NODE 14: Define a function to get 10% of the intensity
    def get_thresh_op(thresh):
        return "-thr %.10f -Tmin -bin" % (0.1 * thresh[1])

    # NODE 15: Determine the median value of the functional runs using the mask
    workflow.connect(maskfunc, "out_file", threshold, "in_file")
    workflow.connect(getthresh, ("out_stat", get_thresh_op), threshold, "op_string")

    # NODE 16: Determine the median value of the functional runs using the mask
    medianval = Node(ImageStats(), name="%s_medianval" % name)
    medianval.long_name = "median value calculation"
    medianval.inputs.op_string = "-k %s -p 50"
    workflow.connect(
        slice_timing_correction, "slice_time_corrected_file", medianval, "in_file"
    )
    workflow.connect(threshold, "out_file", medianval, "mask_file")

    # NODE 17: Dilate the mask
    dilatemask = Node(ImageMaths(), name="%s_dilatemask" % name)
    dilatemask.long_name = "Dilate the mask"
    dilatemask.inputs.suffix = "_dil"
    dilatemask.inputs.op_string = "-dilF"
    workflow.connect(threshold, "out_file", dilatemask, "in_file")

    # NODE 18: Mask the motion corrected functional runs with the dilated mask
    maskfunc2 = Node(ImageMaths(), name="%s_maskfunc2" % name)
    maskfunc2.long_name = "corrected images masking"
    maskfunc2.inputs.suffix = "_mask"
    maskfunc2.inputs.op_string = "-mas"
    workflow.connect(
        slice_timing_correction, "slice_time_corrected_file", maskfunc2, "in_file"
    )
    workflow.connect(dilatemask, "out_file", maskfunc2, "in_file2")

    # NODE 19: Determine the mean image from each functional run
    meanfunc2 = Node(ImageMaths(), name="%s_meanfunc2" % name)
    meanfunc2.long_name = "Mean image calculation"
    meanfunc2.inputs.op_string = "-Tmean"
    meanfunc2.inputs.suffix = "_mean"
    workflow.connect(maskfunc2, "out_file", meanfunc2, "in_file")

    # NODE 20: Merge the median values with the mean functional images into a coupled list
    mergenode = Node(Merge(2), name="%s_mergenode" % name)
    mergenode.long_name = "Mean and median coupling"
    workflow.connect(meanfunc2, "out_file", mergenode, "in1")
    workflow.connect(medianval, "out_stat", mergenode, "in2")

    # NODE 21: Smooth each run using SUSAN with the brightness threshold set to 75% of the
    # median value for each run and a mask constituting the mean functional
    smooth = Node(SUSAN(), name="%s_smooth" % name)
    # Nipype uses a different algorithm to calculate it ->
    # float(fwhm) / np.sqrt(8 * np.log(2)).
    # Therefore, to get 2.12314225053, fwhm should be 4.9996179300001655 instead of 5
    fwhm_thr = 4.9996179300001655
    smooth.inputs.fwhm = fwhm_thr

    # Function to calculate the 75% of the median value
    def get_bt_thresh(medianvals):
        return 0.75 * medianvals

    # Function to define the couple of values
    def get_usans(x):
        return [tuple([x[0], 0.75 * x[1]])]

    workflow.connect(maskfunc2, "out_file", smooth, "in_file")
    workflow.connect(
        medianval, ("out_stat", get_bt_thresh), smooth, "brightness_threshold"
    )
    workflow.connect(mergenode, ("out", get_usans), smooth, "usans")

    # NODE 22: Mask the smoothed data with the dilated mask
    maskfunc3 = Node(ImageMaths(), name="%s_maskfunc3" % name)
    maskfunc3.long_name = "denoised images masking"
    maskfunc3.inputs.suffix = "_mask"
    maskfunc3.inputs.op_string = "-mas"
    workflow.connect(smooth, "smoothed_file", maskfunc3, "in_file")
    workflow.connect(dilatemask, "out_file", maskfunc3, "in_file2")

    # NODE 23: Scale each volume of the run so that the median value of the run is set to 10000
    intnorm = Node(ImageMaths(), name="%s_intnorm" % name)
    intnorm.long_name = "intensity normalization"
    intnorm.inputs.suffix = "_intnorm"

    # Function to get the scaling factor operation string for intensity normalization
    def get_inorm_scale(medianvals):
        return "-mul %.10f" % (10000.0 / medianvals)

    workflow.connect(maskfunc3, "out_file", intnorm, "in_file")
    workflow.connect(medianval, ("out_stat", get_inorm_scale), intnorm, "op_string")

    # NODE 24: Generate a mean functional image from the first run
    meanfunc3 = Node(ImageMaths(), name="%s_meanfunc3" % name)
    meanfunc3.long_name = "mean image calculation"
    meanfunc3.inputs.op_string = "-Tmean"
    meanfunc3.inputs.suffix = "_mean"
    workflow.connect(intnorm, "out_file", meanfunc3, "in_file")

    # NODE 25: Generate the Bunch containing fMRI specifications
    genSpec = Node(FMRIGenSpec(), name="%s_genSpec" % name)
    genSpec.inputs.block_design = block_design
    genSpec.inputs.task_duration = task_duration
    genSpec.inputs.rest_duration = rest_duration
    genSpec.inputs.task_a_name = task_a_name
    genSpec.inputs.task_b_name = task_b_name
    workflow.connect(getTR, "TR", genSpec, "TR")
    workflow.connect(del_vols, "nvols", genSpec, "nvols")
    workflow.connect(meanfunc3, "out_file", genSpec, "tempMean")

    # NODE 26: Perform temporal highpass filtering on the data
    highpass = Node(ImageMaths(), name="%s_highpass" % name)
    highpass.long_name = "Highpass temporal filtering"
    highpass.inputs.suffix = "_tempfilt"
    highpass.inputs.suffix = "_hpf"
    workflow.connect(genSpec, "hpstring", highpass, "op_string")
    workflow.connect(intnorm, "out_file", highpass, "in_file")

    # NODE 27: Coregister the mean functional image to the structural image
    flirt_2_ref = Node(FLIRT(), name="%s_flirt_2_ref" % name)
    flirt_2_ref.long_name = "%s to reference space"
    flirt_2_ref.inputs.out_matrix_file = "fMRI2ref.mat"
    flirt_2_ref.inputs.cost = "corratio"
    flirt_2_ref.inputs.searchr_x = [-90, 90]
    flirt_2_ref.inputs.searchr_y = [-90, 90]
    flirt_2_ref.inputs.searchr_z = [-90, 90]
    flirt_2_ref.inputs.dof = 6
    workflow.connect(meanfunc2, "out_file", flirt_2_ref, "in_file")
    workflow.connect(inputnode, "ref_BET", flirt_2_ref, "reference")

    # NODE 28: Determine which of the images in the functional series are outliers
    # based on deviations in intensity and/or movement.
    art = Node(ArtifactDetect(), name="%s_art" % name)
    art.inputs.use_differences = [True, False]
    art.inputs.use_norm = True
    art.inputs.norm_threshold = 1
    art.inputs.zintensity_threshold = 3
    art.inputs.parameter_source = "FSL"
    art.inputs.mask_type = "file"
    workflow.connect(motion_correct, "par_file", art, "realignment_parameters")
    workflow.connect(motion_correct, "out_file", art, "realigned_files")
    workflow.connect(dilatemask, "out_file", art, "mask_file")

    # NODE 29: Generate design information.
    modelspec = Node(SpecifyModel(), name="%s_modelspec" % name)
    modelspec.inputs.input_units = "secs"
    workflow.connect(genSpec, "hpcutoff", modelspec, "high_pass_filter_cutoff")
    workflow.connect(genSpec, "evs_run", modelspec, "subject_info")
    workflow.connect(getTR, "TR", modelspec, "time_repetition")
    workflow.connect(highpass, "out_file", modelspec, "functional_runs")
    workflow.connect(art, "outlier_files", modelspec, "outlier_files")
    workflow.connect(motion_correct, "par_file", modelspec, "realignment_parameters")

    # NODE 30: Generate a run specific fsf file for analysis
    level_1_design = Node(Level1Design(), name="%s_level_1_design" % name)
    level_1_design.inputs.bases = {"dgamma": {"derivs": False}}
    level_1_design.inputs.model_serial_correlations = True
    workflow.connect(genSpec, "contrasts", level_1_design, "contrasts")
    workflow.connect(getTR, "TR", level_1_design, "interscan_interval")
    workflow.connect(modelspec, "session_info", level_1_design, "session_info")

    # NODE 31: Generate a run specific mat file for use by FILMGLS
    modelgen = Node(FEATModel(), name="%s_modelgen" % name)
    workflow.connect(level_1_design, "fsf_files", modelgen, "fsf_file")
    workflow.connect(level_1_design, "ev_files", modelgen, "ev_files")

    # NODE 32: estimate a model specified by a mat file and a functional run
    modelestimate = Node(FILMGLS(), name="%s_modelestimate" % name)
    modelestimate.inputs.smooth_autocorr = True
    modelestimate.inputs.mask_size = 5
    modelestimate.inputs.threshold = 1000
    workflow.connect(highpass, "out_file", modelestimate, "in_file")
    workflow.connect(modelgen, "design_file", modelestimate, "design_file")
    workflow.connect(modelgen, "con_file", modelestimate, "tcon_file")

    # NODE 33: Get smoothness parameters
    smoothness = Node(SmoothEstimate(), name="%s_smoothness" % name)
    workflow.connect(modelestimate, "residual4d", smoothness, "residual_fit_file")

    # Function to read degree of freedom file
    def dof_from_file(dofFile):
        # Function used out of the box. Import needed
        import os  # TODO trovare modo per sopprimere alert

        if os.path.exists(dofFile):
            with open(dofFile, "r") as file:
                for line in file.readlines():
                    return int(line)

    workflow.connect(modelestimate, ("dof_file", dof_from_file), smoothness, "dof")
    workflow.connect(dilatemask, "out_file", smoothness, "mask_file")

    n_contrasts = 1
    if block_design == BLOCK_DESIGN.RARB:
        n_contrasts += 1
    cont = 0
    while cont < n_contrasts:
        cont += 1

        # NODE 34: Select all result file from filmgls output folder
        results_select = Node(
            SelectFiles(
                {"cope": "cope%d.nii.gz" % cont, "zstat": "zstat%d.nii.gz" % cont}
            ),
            name="%s_results_select_%d" % (name, cont),
        )
        results_select.long_name = "contrast %d result selection" % cont
        workflow.connect(modelestimate, "results_dir", results_select, "base_directory")

        # NODE 35: Mask z-stat with the dilated mask
        maskfunc4 = Node(ImageMaths(), name="%s_maskfunc4_%d" % (name, cont))
        maskfunc4.long_name = "Zstat masking"
        maskfunc4.inputs.suffix = "_mask"
        maskfunc4.inputs.op_string = "-mas"
        workflow.connect(results_select, "zstat", maskfunc4, "in_file")
        workflow.connect(dilatemask, "out_file", maskfunc4, "in_file2")

        # Function to generate the name for the file of output cluster
        def cluster_file_name(contrasts, thres, run_name, x):
            return "r-%s_cluster_%s_threshold%.1f.nii.gz" % (
                run_name,
                contrasts[x - 1][0],
                thres,
            )

        # NODE 36a: Perform clustering on statistical output
        threshold = 3.1
        cluster1 = Node(Cluster(), name="%s_cluster_t3_%d" % (name, cont))
        cluster1.long_name = (
            "contrast "
            + str(cont)
            + " threshold "
            + str(threshold)
            + " %s in reference space"
        )
        cluster1.inputs.threshold = threshold
        cluster1.inputs.connectivity = 26
        cluster1.inputs.pthreshold = 0.05
        cluster1.inputs.out_localmax_txt_file = True

        workflow.connect(
            [
                (
                    genSpec,
                    cluster1,
                    [
                        (
                            ("contrasts", cluster_file_name, threshold, name, cont),
                            "out_threshold_file",
                        )
                    ],
                )
            ]
        )
        workflow.connect(maskfunc4, "out_file", cluster1, "in_file")
        workflow.connect(results_select, "cope", cluster1, "cope_file")
        workflow.connect(smoothness, "volume", cluster1, "volume")
        workflow.connect(smoothness, "dlh", cluster1, "dlh")

        # NODE 37a: Transformation in ref space
        cluster1_2_ref = Node(ApplyXFM(), name="%s_cluster_t3_%d_to_ref" % (name, cont))
        cluster1_2_ref.long_name = (
            "contrast "
            + str(cont)
            + " threshold "
            + str(threshold)
            + " %s in reference space"
        )
        cluster1_2_ref.inputs.apply_xfm = True
        workflow.connect(cluster1, "threshold_file", cluster1_2_ref, "in_file")
        workflow.connect(
            [
                (
                    genSpec,
                    cluster1_2_ref,
                    [
                        (
                            ("contrasts", cluster_file_name, threshold, name, cont),
                            "out_file",
                        )
                    ],
                )
            ]
        )
        workflow.connect(inputnode, "ref_BET", cluster1_2_ref, "reference")
        workflow.connect(
            flirt_2_ref, "out_matrix_file", cluster1_2_ref, "in_matrix_file"
        )

        workflow.connect(
            cluster1_2_ref,
            "out_file",
            outputnode,
            "threshold_file_cont%s_thresh1" % cont,
        )

        # NODE 36b: Perform clustering on statistical output
        threshold = 5
        cluster2 = Node(Cluster(), name="%s_cluster_t5_%d" % (name, cont))
        cluster2.long_name = (
            "contrast "
            + str(cont)
            + " threshold "
            + str(threshold)
            + " %s in reference space"
        )
        cluster2.inputs.threshold = threshold
        cluster2.inputs.connectivity = 26
        cluster2.inputs.pthreshold = 0.05
        cluster2.inputs.out_localmax_txt_file = True

        workflow.connect(
            [
                (
                    genSpec,
                    cluster2,
                    [
                        (
                            ("contrasts", cluster_file_name, threshold, name, cont),
                            "out_threshold_file",
                        )
                    ],
                )
            ]
        )
        workflow.connect(maskfunc4, "out_file", cluster2, "in_file")
        workflow.connect(results_select, "cope", cluster2, "cope_file")
        workflow.connect(smoothness, "volume", cluster2, "volume")
        workflow.connect(smoothness, "dlh", cluster2, "dlh")

        # NODE 37b: Transformation in ref space
        cluster2_2_ref = Node(ApplyXFM(), name="%s_cluster_t5_%d_to_ref" % (name, cont))
        cluster2_2_ref.long_name = (
            "contrast "
            + str(cont)
            + " threshold "
            + str(threshold)
            + " %s in reference space"
        )
        cluster2_2_ref.inputs.apply_xfm = True
        workflow.connect(cluster2, "threshold_file", cluster2_2_ref, "in_file")
        workflow.connect(
            [
                (
                    genSpec,
                    cluster2_2_ref,
                    [
                        (
                            ("contrasts", cluster_file_name, threshold, name, cont),
                            "out_file",
                        )
                    ],
                )
            ]
        )
        workflow.connect(inputnode, "ref_BET", cluster2_2_ref, "reference")
        workflow.connect(
            flirt_2_ref, "out_matrix_file", cluster2_2_ref, "in_matrix_file"
        )

        workflow.connect(
            cluster2_2_ref,
            "out_file",
            outputnode,
            "threshold_file_cont%s_thresh2" % cont,
        )

        # NODE 36c: Perform clustering on statistical output
        threshold = 7
        cluster3 = Node(Cluster(), name="%s_cluster_t7_%d" % (name, cont))
        cluster3.long_name = (
            "contrast "
            + str(cont)
            + " threshold "
            + str(threshold)
            + " %s in reference space"
        )
        cluster3.inputs.threshold = threshold
        cluster3.inputs.connectivity = 26
        cluster3.inputs.pthreshold = 0.05
        cluster3.inputs.out_localmax_txt_file = True

        workflow.connect(
            [
                (
                    genSpec,
                    cluster3,
                    [
                        (
                            ("contrasts", cluster_file_name, threshold, name, cont),
                            "out_threshold_file",
                        )
                    ],
                )
            ]
        )
        workflow.connect(maskfunc4, "out_file", cluster3, "in_file")
        workflow.connect(results_select, "cope", cluster3, "cope_file")
        workflow.connect(smoothness, "volume", cluster3, "volume")
        workflow.connect(smoothness, "dlh", cluster3, "dlh")

        # NODE 37c: Transformation in ref space
        cluster3_2_ref = Node(ApplyXFM(), name="%s_cluster_t7_%d_to_ref" % (name, cont))
        cluster3_2_ref.long_name = (
            "contrast "
            + str(cont)
            + " threshold "
            + str(threshold)
            + " %s in reference space"
        )
        cluster3_2_ref.inputs.apply_xfm = True
        workflow.connect(cluster3, "threshold_file", cluster3_2_ref, "in_file")
        workflow.connect(
            [
                (
                    genSpec,
                    cluster3_2_ref,
                    [
                        (
                            ("contrasts", cluster_file_name, threshold, name, cont),
                            "out_file",
                        )
                    ],
                )
            ]
        )
        workflow.connect(inputnode, "ref_BET", cluster3_2_ref, "reference")
        workflow.connect(
            flirt_2_ref, "out_matrix_file", cluster3_2_ref, "in_matrix_file"
        )

        workflow.connect(
            cluster3_2_ref,
            "out_file",
            outputnode,
            "threshold_file_cont%s_thresh3" % cont,
        )

    return workflow



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from os.path import abspath
from nipype.interfaces.base import (
    BaseInterface,
    BaseInterfaceInputSpec,
    TraitedSpec,
    InputMultiPath,
    File,
    isdefined,
)


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterfaceInputSpec)  -*-
class MergeTargetsInputSpec(BaseInterfaceInputSpec):
    target_files = InputMultiPath(
        File(exists=True), mandatory=True, desc="list of path file to merge in txt"
    )
    out_file = File(desc="the output file name")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class MergeTargetsOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output txt file")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterface)  -*-
class MergeTargets(BaseInterface):
    """
    Creates a .txt file from an array.

    """

    input_spec = MergeTargetsInputSpec
    output_spec = MergeTargetsOutputSpec

    def _run_interface(self, runtime):
        self.inputs.out_file = self._gen_outfilename()

        lines = "\n".join(self.inputs.target_files)

        with open(self.inputs.out_file, "w") as file:
            file.write(lines)

        return runtime

    def _gen_outfilename(self):
        out_file = self.inputs.out_file
        if not isdefined(out_file):
            out_file = "targets.txt"
        return abspath(out_file)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self._gen_outfilename()
        return outputs



import sys
import os
from datetime import datetime
from swane import strings
from inspect import getframeinfo, stack

ERROR_FILE = os.path.abspath(
    os.path.join(os.path.expanduser("~"), "." + strings.APPNAME) + "_errorlog"
)


def print_error():
    """
    Print a standardized Error Log string

    Parameters
    ----------
    e : Exception
        The Exception TypeError.

    Returns
    -------
    string
        The formatted error string.

    """

    try:
        exception_type = repr(sys.exc_info()[1])
        caller = getframeinfo(stack()[1][0])
        function_name = caller.function
        file_name = caller.filename
        line_number = caller.lineno
        message = f"\n\n{datetime.now().strftime('%Y/%m/%d, %H:%M:%S')} - File name: {file_name} - Func name: {function_name} - Exception type: {exception_type} at Line: {line_number}"

        with open(ERROR_FILE, "a+") as f:
            f.write(message)

    except Exception as e:
        exception_type, exception_object, exception_traceback = sys.exc_info()
        line_number = exception_traceback.tb_lineno
        print(
            f"{datetime.now().strftime('%Y/%m/%d, %H:%M:%S')} - "
            + f"File name: {os.path.normpath(os.path.basename(exception_traceback.tb_frame.f_code.co_filename))} - Func name: print_error - "
            + f"Exception type: {exception_type} "
            + f"at Line: {line_number} - "
            + f"{str(e)}"
        )



import smtplib
from datetime import datetime
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart


class MailManager:
    """
    Mail Sending service manager
    """

    TIMEOUT = 2

    def __init__(
        self,
        server_address: str,
        server_port: int,
        username: str,
        password: str,
        use_ssl: bool = True,
        use_tls: bool = False,
    ):
        """
        The MailManager initializer.

        Parameters
        ----------
        server_address : str
            The mail host server address
        server_port : int
            The mail host server port
        username : str
            The email
        password : str
            The email password
        use_ssl : bool. Optional
            True to use the SSL protocol, otherwhise False. The default is True
        use_tls : bool. Optional
            True to use the TLS protocol, otherwhise False. Useless if use_ssl is True. The default is False

        """

        self.server_address = server_address
        self.server_port = server_port
        self.username = username
        self.password = password
        self.use_ssl = use_ssl
        self.use_tls = use_tls
        self.server = None

    def connect(self):
        """
        Connects to the SMTP server using the provided credentials.

        """

        try:
            if self.use_ssl:
                self.server = smtplib.SMTP_SSL(
                    self.server_address, self.server_port, timeout=MailManager.TIMEOUT
                )
            else:
                self.server = smtplib.SMTP(
                    self.server_address, self.server_port, timeout=MailManager.TIMEOUT
                )
                self.server.ehlo()
                if self.use_tls:
                    self.server.starttls()
                    self.server.ehlo()
            self.server.login(self.username, self.password)
        except Exception as e:
            raise Exception(f"{str(e)} - Check your Mail Configuration")

    def disconnect(self):
        """
        Disconnects from the SMTP server.

        """

        if self.server:
            self.server.quit()

    def send_mail(self, from_addr, to_addr, subject, body):
        """
        Sends an email with the specified subject and body.

        """

        try:
            message = MIMEMultipart()
            message["From"] = from_addr
            message["To"] = to_addr
            message["Subject"] = subject
            message.attach(MIMEText(body, "html"))

            self.connect()
            self.server.send_message(message)
            self.disconnect()
        except Exception as e:
            raise

    def send_report(self, body):
        """
        Sends a report for a SWANe workflow based on the user mail configuration
        """

        self.send_mail(self.username, self.username, f"SWANe - {datetime.now()}", body)



import os
from typing import Union
from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QLabel,
    QLineEdit,
    QPushButton,
    QFileDialog,
    QMessageBox,
    QCheckBox,
    QSpinBox,
    QDoubleSpinBox,
    QComboBox,
    QStyle,
    QSizePolicy,
    QStyleOption,
    QWidget,
)
from functools import partial
from enum import Enum
from swane import strings
from swane.config.config_enums import InputTypes
from swane.config.ConfigManager import ConfigManager
from swane.config.PreferenceEntry import PreferenceEntry
from swane.utils.CryptographyManager import CryptographyManager


class PreferenceUIEntry:
    """
    This class manage The label and the input of a preference based on the value of a PreferenceEntry
    """

    def __init__(
        self,
        category: Enum,
        key: str,
        my_config: ConfigManager,
        entry: PreferenceEntry,
        parent=None,
    ):

        # Var initialization
        self.changed = False
        self.category = category
        self.key = key
        self.parent = parent
        self.label = QLabel()
        self.tooltip = None
        opt = QStyleOption()
        opt.initFrom(self.label)
        text_size = self.label.fontMetrics().size(
            Qt.TextShowMnemonic, self.label.text()
        )
        height = (
            self.label.style()
            .sizeFromContents(QStyle.CT_PushButton, opt, text_size, self.label)
            .height()
        )
        self.label.setMaximumHeight(height)
        self.label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.box_text = ""

        # Import values from entry
        self.input_type = entry.input_type
        self.restart = entry.restart
        self.informative_text = entry.informative_text
        self.input_field, self.button = self.gen_input_field()
        self.informative_text_label = self.gen_informative_text_label()
        if self.input_type == InputTypes.ENUM and entry.value_enum is not None:
            self.populate_combo(entry.value_enum)
        self.validate_on_change = entry.validate_on_change

        # Apply values to GUI elements
        self.set_label_text(entry.label)
        self.set_tooltip(entry.tooltip)
        self.set_informative_text()
        if entry.range is not None:
            self.set_range(entry.range[0], entry.range[1])
        if entry.decimals is not None:
            self.set_decimals(entry.decimals)
        self.set_value_from_config(my_config)

        # Change-connect must be the last step, otherwise it is fired during initialization
        self.connect_change(self.set_changed)

    def set_label_text(self, text: str):
        self.label.setText(text)

    def set_box_text(self, text: str):
        self.box_text = text

    def set_changed(self, **kwargs):
        """
        Fired when the input value is changed.
        Used for config saving and eventual application restart
        """
        self.changed = True
        if self.restart and self.parent is not None:
            self.parent.set_restart()

    def connect_change(self, callback_function: callable):
        """
        Connect the input field to a callback function

        Parameters
        ----------
        callback_function: callable
            The function to connect
        """
        if callback_function is None:
            return
        if self.input_type == InputTypes.BOOLEAN:
            self.input_field.toggled.connect(callback_function)
        elif self.input_type == InputTypes.ENUM:
            self.input_field.currentIndexChanged.connect(callback_function)
        else:
            self.input_field.textChanged.connect(callback_function)

    def gen_informative_text_label(self) -> QLabel:
        """
        Generate a QLabel to display informative text related to current preference value

        Return
        ----------
        The QLabel
        """
        informative_text_label = None
        if self.informative_text is not None:
            informative_text_label = QLabel()
            self.connect_change(self.set_informative_text)
        return informative_text_label

    def set_informative_text(self, **kwargs):
        """
        Update the informative_text_label when the preference value is changed
        """
        if self.informative_text_label is None or self.informative_text is None:
            return
        if self.get_typed_value() not in self.informative_text:
            return
        self.informative_text_label.setText(
            self.informative_text[self.get_typed_value()]
        )

    def gen_input_field(self) -> [QWidget, QPushButton]:
        """
        Generate an input field using a widget based on preference type and, if required, a button to open a file selector

        Return
        ----------
        The input field widget and the button
        """
        button = None

        if self.input_type == InputTypes.BOOLEAN:
            field = QCheckBox()
        elif self.input_type == InputTypes.ENUM:
            field = QComboBox()
        elif self.input_type == InputTypes.INT:
            field = QSpinBox()
            field.setMinimum(-1)
            field.setMaximum(100)
        elif self.input_type == InputTypes.FLOAT:
            field = QDoubleSpinBox()
            field.setMinimum(0)
            field.setMaximum(100)
            field.setDecimals(2)
            field.setSingleStep(0.01)
        else:
            field = QLineEdit()

        if (
            self.input_type == InputTypes.FILE
            or self.input_type == InputTypes.DIRECTORY
        ):
            field.setReadOnly(True)
            button = QPushButton()
            pixmap = getattr(QStyle, "SP_DirOpenIcon")
            icon_open_dir = button.style().standardIcon(pixmap)
            button.setIcon(icon_open_dir)
            button.clicked.connect(self.choose_file)

        if self.input_type == InputTypes.PASSWORD:
            field.setEchoMode(QLineEdit.PasswordEchoOnEdit)
            field.editingFinished.connect(partial(self.encrypt_password, field))

        return field, button

    @staticmethod
    def encrypt_password(field: QLineEdit):
        field.setText(CryptographyManager.encrypt(field.text()))

    def choose_file(self):
        """
        Open a selection window to chose a file ora directory
        """
        if self.input_type == InputTypes.FILE:
            file_path, _ = QFileDialog.getOpenFileName(
                parent=self.parent, caption=self.box_text
            )
            error = strings.pref_window_file_error
        elif self.input_type == InputTypes.DIRECTORY:
            file_path = QFileDialog.getExistingDirectory(
                parent=self.parent, caption=self.box_text
            )
            error = strings.pref_window_dir_error
        else:
            return

        if file_path == "":
            return

        if not os.path.exists(file_path):
            msg_box = QMessageBox()
            msg_box.setIcon(QMessageBox.Warning)
            msg_box.setText(error)
            msg_box.exec()
            return

        if self.validate_on_change:
            file_path = "*" + file_path

        self.set_value(file_path)

    def populate_combo(self, items: Enum):
        """
        In case of QComboBox, populates it using an Enum members.
        Each Enum member is stored in menu Data and it's value is used as label

        Parameters
        ----------
        items: Enum
            The Enum iterated to populate the ComboBox

        """
        if self.input_type != InputTypes.ENUM or items is None:
            return
        for member in items:
            self.input_field.addItem(member.value, userData=member)

    def set_value_from_config(self, config: ConfigManager):
        """
        Get the value of a preference forcing it's type based on self.input_type

        Parameters
        ----------
        config: ConfigManager
           The configuration from which read the value

        """
        if config is None:
            return

        if self.input_type == InputTypes.BOOLEAN:
            value = config.getboolean_safe(self.category, self.key)
        elif self.input_type == InputTypes.INT:
            value = config.getint_safe(self.category, self.key)
        elif self.input_type == InputTypes.ENUM:
            value = config.getenum_safe(self.category, self.key)
        elif self.input_type == InputTypes.FLOAT:
            value = config.getfloat_safe(self.category, self.key)
        else:
            value = config[self.category][self.key]

        self.set_value(value)

        # Convert Enum to their label for compare reason
        if hasattr(value, "name"):
            value = value.name
        if str(value).lower() != config[self.category][self.key].lower():
            self.set_changed()

    def set_range(self, min_value: Union[int, float], max_value: Union[int, float]):
        """
        Apply specified values to a numeric input field validator

        Parameters
        ----------
        min_value: int or float
            The minimum accepted value
        max_value: int or float
           The maximum accepted value
        """

        if not isinstance(self.input_field, QSpinBox | QDoubleSpinBox):
            return

        if min_value > max_value:
            min_value, max_value = max_value, min_value

        self.input_field.setMinimum(min_value)
        self.input_field.setMaximum(max_value)

    def set_decimals(self, decimals: int):
        """
        Apply specified values to a numeric input field validator

        Parameters
        ----------
        decimal: int
            The number of decimals accepted by the field
        """

        if not isinstance(self.input_field, QDoubleSpinBox):
            return

        single_step = 10**-decimals

        self.input_field.setDecimals(decimals)
        self.input_field.setSingleStep(single_step)

    def set_value(self, value, reset_change_state: bool = False):
        """
        Insert a value in the input field

        Parameters
        ----------
        value
            The new value to be displayed
        reset_change_state: bool, optional
           If True, the field change status will not be set to True. Default is False.
        """
        if self.input_type == InputTypes.BOOLEAN:
            if type(value) is not bool:
                raise Exception("Non boolean value for checkbox: %s" % str(value))
            if value:
                self.input_field.setCheckState(Qt.Checked)
            else:
                self.input_field.setCheckState(Qt.Unchecked)
        elif self.input_type == InputTypes.ENUM:
            if self.input_field.count() == 0:
                return
            index = self.input_field.findData(value)
            self.input_field.setCurrentIndex(index)
        elif self.input_type == InputTypes.INT:
            self.input_field.setValue(int(value))
        elif self.input_type == InputTypes.FLOAT:
            self.input_field.setValue(float(value))
        else:
            self.input_field.setText(str(value))

        if reset_change_state:
            self.changed = False

    def disable(self, tooltip: str = None):
        """
        Disable the input field and gray out the label

        Parameters
        ----------
        tooltip: str, optional
            A tooltip to be displayed on label and input field. Default is None
        """
        self.input_field.setEnabled(False)
        self.input_field.setStyleSheet("color: gray")
        self.label.setStyleSheet("color: gray")
        self.set_tooltip(tooltip)
        if self.input_type == InputTypes.BOOLEAN:
            self.input_field.setChecked(False)

    def set_tooltip(self, tooltip: str):
        """
        Display a tooltip on label and input field

        Parameters
        ----------
        tooltip: str
            The tooltip string
        """
        if self.tooltip is None:
            self.tooltip = tooltip
        if tooltip == "" and self.tooltip != "":
            tooltip = self.tooltip
        self.input_field.setToolTip(tooltip)
        self.label.setToolTip(tooltip)
        if tooltip == "":
            self.label.setText(self.label.text().replace(" " + strings.INFOCHAR, ""))
        elif not self.label.text().endswith(strings.INFOCHAR):
            self.label.setText(self.label.text() + " " + strings.INFOCHAR)

    def enable(self):
        """
        Enable the input field and remove gray out from the label
        """
        self.input_field.setEnabled(True)
        self.input_field.setStyleSheet("")
        self.set_tooltip(self.tooltip)
        self.label.setStyleSheet("")

    def get_value(self) -> str:
        """
        Return
        ----------
        The value of the input field as a string
        """
        if self.input_type == InputTypes.ENUM:
            value = self.input_field.itemData(self.input_field.currentIndex()).name
        elif self.input_type == InputTypes.BOOLEAN:
            if self.input_field.checkState() == Qt.Checked:
                value = "true"
            else:
                value = "false"
        elif self.input_type == InputTypes.FLOAT or self.input_type == InputTypes.INT:
            value = str(self.input_field.value())
        else:
            value = self.input_field.text()

        return value

    def get_typed_value(self):
        """
        Return
        ----------
        The value of the input field as a boolean, Enum or string
        """
        if self.input_type == InputTypes.ENUM:
            value = self.input_field.itemData(self.input_field.currentIndex())
        elif self.input_type == InputTypes.BOOLEAN:
            return self.input_field.checkState() == Qt.Checked
        else:
            value = self.input_field.text()

        return value



import subprocess
from sys import platform
import os


def open_results_directory(pushed: bool, results_dir: str):
    """
    Open the result directory in the file explorer
    Parameters
    ----------
    pushed: bool
        Unused, passed by QPushButton
    results_dir: str
        The path to show in file explorer
    """
    if platform == "win32":
        os.startfile(results_dir)
    elif platform == "darwin":
        subprocess.Popen(["open", results_dir])

    else:
        try:
            subprocess.Popen(["xdg-open", results_dir])
        except OSError:
            pass



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

from nipype.interfaces.fsl.base import FSLCommand, FSLCommandInputSpec
from nipype.interfaces.base import traits, TraitedSpec, File, isdefined


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.fsl.base.FSLCommandInputSpec)  -*-
class OrientInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True, mandatory=True, argstr="%s", position="2", desc="input image"
    )
    _options_xor = ["get_orient", "swap_orient"]
    get_orient = traits.Bool(
        argstr="-getorient",
        position="1",
        xor=_options_xor,
        desc="gets FSL left-right orientation",
    )
    swap_orient = traits.Bool(
        argstr="-swaporient",
        position="1",
        xor=_options_xor,
        desc="swaps FSL radiological and FSL neurological",
    )


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class OrientOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="image with modified orientation")
    orient = traits.Str(desc="FSL left-right orientation")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.fsl.base.FSLCommand)  -*-
class Orient(FSLCommand):
    """
    Returns the image orientation as neurological or radiological conventions.

    """

    _cmd = "fslorient"
    input_spec = OrientInputSpec
    output_spec = OrientOutputSpec

    def aggregate_outputs(self, runtime=None, needed_outputs=None):

        outputs = self._outputs()
        info = runtime.stdout

        # Modified file
        if isdefined(self.inputs.swap_orient):
            outputs.out_file = self.inputs.in_file

        # Get information
        if isdefined(self.inputs.get_orient):
            outputs.orient = info

        return outputs



import swane_supplement

from nipype.interfaces.fsl import (
    ApplyWarp,
    ApplyMask,
    BinaryMaths,
    FAST,
    ImageStats,
    SpatialFilter,
    Threshold,
)
from nipype.pipeline.engine import Node

from swane.nipype_pipeline.nodes.utils import getn
from swane.nipype_pipeline.engine.CustomWorkflow import CustomWorkflow
from swane.nipype_pipeline.nodes.ThrROI import ThrROI
from swane.nipype_pipeline.nodes.FLAT1OutliersMask import FLAT1OutliersMask

from nipype.interfaces.utility import IdentityInterface


def flat1_workflow(name: str, mni1_dir: str, base_dir: str = "/") -> CustomWorkflow:
    """
    Creation of a junction and extension z-score map based on T13D, FLAIR3D and
    a mean template.

    Parameters
    ----------
    name : str
        The workflow name.
    mni1_dir : path
        The file path of the MNI1 template.
    base_dir : path, optional
        The base directory path relative to parent workflow. The default is "/".

    Input Node Fields
    ----------
    ref_brain : path
        Betted T13D reference file.
    flair_brain : path
        Betted FLAIR3D.
    ref_2_mni1_warp : path
        Nonlinear registration warp from T13D reference space to MNI1 atlas.
    ref_2_mni1_inverse_warp : path
        Nonlinear registration inverse warp from MNI1 atlas space to T13D reference.

    Returns
    -------
    workflow : CustomWorkflow
        The FLAT1 workflow.

    Output Node Fields
    ----------
    extension_z : path
        Extension z-score map in T13D reference space.
        Based on:
        - Pascher et al. - Automated morphometric MRI analysis for the detection of PNH - Epilepsia 2013
    junction_z : path
        Junction z-score map in T13D reference space.
        Based on:
        - Huppertz et al. - Enhanced visualization of blurred gray-white matter junctions in focal cortical
                            dysplasia by voxel-based 3D MRI analysis - Epilepsy Res 2005
        - Huppertz et al. - Voxel-based 3D MRI analysis helps to detect subtle forms of subcortical band
                            heterotopia - Epilepsia 2008
    binary_flair : path
        Divided image FLAIR/T13D.

    """

    workflow = CustomWorkflow(name=name, base_dir=base_dir)

    # Input Node
    inputnode = Node(
        IdentityInterface(
            fields=[
                "ref_brain",
                "flair_brain",
                "ref_2_mni1_warp",
                "ref_2_mni1_inverse_warp",
            ]
        ),
        name="inputnode",
    )

    # Output Node
    outputnode = Node(
        IdentityInterface(fields=["extension_z", "junction_z", "binary_flair"]),
        name="outputnode",
    )

    # NODE 1: three class fast segmentation
    fast = Node(FAST(), name="%s_fast" % name)
    fast.inputs.img_type = 1  # param -t
    fast.inputs.number_classes = 3  # param n
    fast.inputs.hyper = 0.1  # param -H
    fast.inputs.bias_lowpass = 40  # param -l
    fast.inputs.output_biascorrected = True  # param -B
    fast.inputs.bias_iters = 4  # param -I
    workflow.add_nodes([fast])
    workflow.connect(inputnode, "ref_brain", fast, "in_files")

    # NODE 2: FLAIR nonlinear transformation in MNI1 atlas space
    flair_2_mni1 = Node(ApplyWarp(), name="%s_flair2mni1" % name)
    flair_2_mni1.long_name = "Flair %s to MNI space"
    flair_2_mni1.inputs.ref_file = mni1_dir
    workflow.add_nodes([flair_2_mni1])
    workflow.connect(inputnode, "flair_brain", flair_2_mni1, "in_file")
    workflow.connect(inputnode, "ref_2_mni1_warp", flair_2_mni1, "field_file")

    # NODE 3: t1_restore nonlinear transformation in MNI1 atlas space
    restore_2_mni1 = Node(ApplyWarp(), name="%s_restore2mni1" % name)
    restore_2_mni1.long_name = "T1 %s to MNI space"
    restore_2_mni1.inputs.ref_file = mni1_dir
    workflow.connect(fast, "restored_image", restore_2_mni1, "in_file")
    workflow.connect(inputnode, "ref_2_mni1_warp", restore_2_mni1, "field_file")

    # NODE 4: Gray matter nonlinear transformation in MNI1 atlas space
    gm_2_mni1 = Node(ApplyWarp(), name="%s_gm2mni1" % name)
    gm_2_mni1.long_name = "Grey matter %s to MNI space"
    gm_2_mni1.inputs.ref_file = mni1_dir
    workflow.connect(
        [(fast, gm_2_mni1, [(("partial_volume_files", getn, 1), "in_file")])]
    )
    workflow.connect(inputnode, "ref_2_mni1_warp", gm_2_mni1, "field_file")

    # NODE 5: White matter nonlinear transformation in MNI1 atlas space
    wm_2_mni1 = Node(ApplyWarp(), name="%s_wm2mni1" % name)
    wm_2_mni1.long_name = "White matter %s to MNI space"
    wm_2_mni1.inputs.ref_file = mni1_dir
    workflow.connect(
        [(fast, wm_2_mni1, [(("partial_volume_files", getn, 2), "in_file")])]
    )
    workflow.connect(inputnode, "ref_2_mni1_warp", wm_2_mni1, "field_file")

    # NODE 6: Divided image generation from FLAIR/T1
    flair_div_ref = Node(BinaryMaths(), name="%s_flairDIVref" % name)
    flair_div_ref.long_name = "Flair/T1 normalization"
    flair_div_ref.inputs.operation = "div"
    workflow.connect(flair_2_mni1, "out_file", flair_div_ref, "in_file")
    workflow.connect(restore_2_mni1, "out_file", flair_div_ref, "operand_file")

    # Skip this step, sometimes causes wrong cortical segmentation!
    # NODE 7: Outliers removal from mask
    # outliers_mask = Node(FLAT1OutliersMask(), name="%s_outliers_mask" % name)
    # outliers_mask.inputs.mask_file = swane_supplement.cortex_mas
    # workflow.connect(flair_div_ref, "out_file", outliers_mask, "in_file")

    # Remove the upper 1% of values to trim values from incorrect registration
    outliers_removal = Node(Threshold(), name="%s_outliers_mask" % name)
    outliers_removal.inputs.thresh = 98
    outliers_removal.inputs.use_robust_range = True
    outliers_removal.inputs.use_nonzero_voxels = True
    outliers_removal.inputs.direction = "above"
    workflow.connect(flair_div_ref, "out_file", outliers_removal, "in_file")

    # NODE 8: Cerebellum removal from divided image
    cortex_mask = Node(ApplyMask(), name="%s_cortexMask" % name)
    cortex_mask.long_name = "outliers %s"
    cortex_mask.inputs.mask_file = swane_supplement.cortex_mas
    # workflow.connect(outliers_mask, "out_file", cortex_mask, "mask_file")
    # workflow.connect(flair_div_ref, "out_file", cortex_mask, "in_file")
    workflow.connect(outliers_removal, "out_file", cortex_mask, "in_file")

    # NODE 9: Masking for gray matter on t1_restore in MNI1
    gm_mask = Node(ApplyMask(), name="%s_gmMask" % name)
    gm_mask.long_name = "Grey matter %s"
    workflow.connect(cortex_mask, "out_file", gm_mask, "in_file")
    workflow.connect(gm_2_mni1, "out_file", gm_mask, "mask_file")

    # NODE 10: Masking for white matter on t1_restore in MNI1
    wm_mask = Node(ApplyMask(), name="%s_wmMask" % name)
    wm_mask.long_name = "White matter %s"
    workflow.connect(cortex_mask, "out_file", wm_mask, "in_file")
    workflow.connect(wm_2_mni1, "out_file", wm_mask, "mask_file")

    # NODE 11: Mean calculation for gray matter
    gm_mean = Node(ImageStats(), name="%s_gm_mean" % name)
    gm_mean.long_name = "Grey matter mean value calculation"
    gm_mean.inputs.op_string = "-M"
    workflow.connect(gm_mask, "out_file", gm_mean, "in_file")

    # NODE 12: Mean calculation for white matter
    wm_mean = Node(ImageStats(), name="%s_wm_mean" % name)
    wm_mean.long_name = "White matter mean value calculation"
    wm_mean.inputs.op_string = "-M"
    workflow.connect(wm_mask, "out_file", wm_mean, "in_file")

    # TODO parametri per ora inutilizzati. Valutare in futuro la loro implementazione
    # FLAT1_gm_std = Node(ImageStats(), name="%s_gm_std")
    # FLAT1_gm_std.inputs.op_string="-S"
    # workflow.connect(FLAT1_gmMask,"out_file",FLAT1_gm_std,"in_file")

    # FLAT1_wm_std = Node(ImageStats(), name="%s_wm_std")
    # FLAT1_wm_std.inputs.op_string="-S"
    # workflow.connect(FLAT1_wmMask,"out_file",FLAT1_wm_std,"in_file")

    # NODE 13: Mask generation with values between mean white matter and mean gray matter values
    binary_flair = Node(ThrROI(), name="%s_binaryFLAIR" % name)
    binary_flair.long_name = "Mean based masking"
    binary_flair.inputs.out_file = "binary_flair.nii.gz"
    workflow.connect(cortex_mask, "out_file", binary_flair, "in_file")
    workflow.connect(gm_mean, "out_stat", binary_flair, "seg_val_max")
    workflow.connect(wm_mean, "out_stat", binary_flair, "seg_val_min")

    # NODE 14: Junction map generation
    convolution_flair = Node(SpatialFilter(), name="%s_convolution_flair" % name)
    convolution_flair.long_name = "junction map generation"
    convolution_flair.inputs.operation = "mean"  # Param -fmean
    convolution_flair.inputs.kernel_shape = "boxv"  # Param -kernel
    convolution_flair.inputs.kernel_size = 5  # Param -kernel value
    convolution_flair.inputs.out_file = "convolution_flair.nii.gz"
    workflow.connect(binary_flair, "out_file", convolution_flair, "in_file")

    # NODE 13: Junction map mean value calculation
    junction_mean = Node(BinaryMaths(), name="%s_junction_mean" % name)
    junction_mean.long_name = "junction variation from mean atlas"
    junction_mean.inputs.operation = "sub"  # Param -sub
    junction_mean.inputs.operand_file = swane_supplement.mean_flair
    junction_mean.inputs.out_file = "junction_flair.nii.gz"
    workflow.connect(convolution_flair, "out_file", junction_mean, "in_file")

    # NODE 14: Junction z-score calculation
    junction_z = Node(BinaryMaths(), name="%s_junctionz" % name)
    junction_z.long_name = "junction z score calculation"
    junction_z.inputs.operation = "div"
    junction_z.inputs.operand_file = swane_supplement.std_final_flair
    junction_z.inputs.out_file = "junctionZ_flair.nii.gz"
    workflow.connect(junction_mean, "out_file", junction_z, "in_file")

    # NODE 15: Cerebellum mask on restore_t1
    masked_cerebellum = Node(ApplyMask(), name="%s_masked_cerebellum" % name)
    masked_cerebellum.long_name = "cerebellum %s"
    masked_cerebellum.inputs.mask_file = swane_supplement.binary_cerebellum
    workflow.connect(restore_2_mni1, "out_file", masked_cerebellum, "in_file")

    # NODE 16: Cerebellum mean value calculation
    cerebellum_mean = Node(ImageStats(), name="%s_cerebellum_mean" % name)
    cerebellum_mean.long_name = "cerebellum mean value calculation"
    cerebellum_mean.inputs.op_string = "-M"
    workflow.connect(masked_cerebellum, "out_file", cerebellum_mean, "in_file")

    # NODE 17: Grey matter mask on restore_t1
    restore_gm_mask = Node(ApplyMask(), name="%s_restore_gmMask" % name)
    restore_gm_mask.long_name = "grey matter %s"
    restore_gm_mask.inputs.out_file = "masked_image_GM.nii.gz"
    workflow.connect(restore_2_mni1, "out_file", restore_gm_mask, "in_file")
    workflow.connect(gm_2_mni1, "out_file", restore_gm_mask, "mask_file")

    # NODE 18: Grey matter normalization on cerebellum mean value
    normalised_gm_mask = Node(BinaryMaths(), name="%s_normalised_GM_mask" % name)
    normalised_gm_mask.long_name = "Grey matter/cerebellum normalization"
    normalised_gm_mask.inputs.operation = "div"
    normalised_gm_mask.inputs.out_file = "normalised_GM_mask.nii.gz"
    workflow.connect(restore_gm_mask, "out_file", normalised_gm_mask, "in_file")
    workflow.connect(cerebellum_mean, "out_stat", normalised_gm_mask, "operand_value")

    # NODE 19: Extension map generation
    smoothed_image_extension = Node(
        SpatialFilter(), name="%s_smoothed_image_extension" % name
    )
    smoothed_image_extension.long_name = "extension map generation"
    smoothed_image_extension.inputs.operation = "mean"  # Param -fmean
    smoothed_image_extension.inputs.kernel_shape = "boxv"  # Param -kernel
    smoothed_image_extension.inputs.kernel_size = 5  # Param -kernel value
    smoothed_image_extension.inputs.out_file = "smoothed_image_extension.nii.gz"
    workflow.connect(
        normalised_gm_mask, "out_file", smoothed_image_extension, "in_file"
    )

    # NODE 20: Extension map mean value calculation
    extension_mean = Node(BinaryMaths(), name="%s_image_extension" % name)
    extension_mean.long_name = "extension variation from mean atlas"
    extension_mean.inputs.operation = "sub"
    extension_mean.inputs.operand_file = swane_supplement.mean_extension
    extension_mean.inputs.out_file = "extension_image.nii.gz"
    workflow.connect(smoothed_image_extension, "out_file", extension_mean, "in_file")

    # NODE 21: Extension z-score calculation
    extension_z = Node(BinaryMaths(), name="%s_image_extensionz" % name)
    extension_z.long_name = "extension z score calculation"
    extension_z.inputs.operation = "div"
    extension_z.inputs.operand_file = swane_supplement.std_final_extension
    extension_z.inputs.out_file = "extension_z.nii.gz"
    workflow.connect(extension_mean, "out_file", extension_z, "in_file")

    # NODE 22: Cerebellum removal from extension z-score map
    no_cereb_extension_z = Node(ApplyMask(), name="%s_no_cereb_extension_z" % name)
    no_cereb_extension_z.long_name = "cerebellum %s"
    no_cereb_extension_z.inputs.out_file = "no_cereb_extension_z.nii.gz"
    workflow.connect(extension_z, "out_file", no_cereb_extension_z, "in_file")
    # workflow.connect(outliers_mask, "out_file", no_cereb_extension_z, "mask_file")
    no_cereb_extension_z.inputs.mask_file = swane_supplement.cortex_mas

    # NODE 23: Extension z-score nonlinear transformation in reference space
    extension_z_2_ref = Node(ApplyWarp(), name="%s_extensionz2ref" % name)
    extension_z_2_ref.long_name = "extension %s to reference space"
    extension_z_2_ref.inputs.out_file = "r-extension_z.nii.gz"
    workflow.connect(no_cereb_extension_z, "out_file", extension_z_2_ref, "in_file")
    workflow.connect(
        inputnode, "ref_2_mni1_inverse_warp", extension_z_2_ref, "field_file"
    )
    workflow.connect(inputnode, "ref_brain", extension_z_2_ref, "ref_file")

    # NODE 24: Junction z-score nonlinear transformation in reference space
    junction_z_2_ref = Node(ApplyWarp(), name="%s_junctionz2ref" % name)
    junction_z_2_ref.long_name = "junction %s to reference space"
    junction_z_2_ref.inputs.out_file = "r-junction_z.nii.gz"
    workflow.connect(junction_z, "out_file", junction_z_2_ref, "in_file")
    workflow.connect(
        inputnode, "ref_2_mni1_inverse_warp", junction_z_2_ref, "field_file"
    )
    workflow.connect(inputnode, "ref_brain", junction_z_2_ref, "ref_file")

    # NODE 25: Divided image nonlinear transformation in reference space
    binary_flair_2_ref = Node(ApplyWarp(), name="%s_binaryFLAIR2ref" % name)
    binary_flair_2_ref.long_name = "Normalized Flair %s to reference space"
    binary_flair_2_ref.inputs.out_file = "r-binary_flair.nii.gz"
    workflow.connect(binary_flair, "out_file", binary_flair_2_ref, "in_file")
    workflow.connect(
        inputnode, "ref_2_mni1_inverse_warp", binary_flair_2_ref, "field_file"
    )
    workflow.connect(inputnode, "ref_brain", binary_flair_2_ref, "ref_file")

    workflow.connect(extension_z_2_ref, "out_file", outputnode, "extension_z")
    workflow.connect(junction_z_2_ref, "out_file", outputnode, "junction_z")
    workflow.connect(binary_flair_2_ref, "out_file", outputnode, "binary_flair")

    return workflow



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

import traceback
from nipype.pipeline.plugins.multiproc import MultiProcPlugin
from swane.nipype_pipeline.engine.WorkflowReport import WorkflowReport, WorkflowSignals
from swane import strings


# -*- DISCLAIMER: this class extends a Nipype class (nipype.pipeline.plugins.multiproc.MultiProcPlugin)  -*-
class MonitoredMultiProcPlugin(MultiProcPlugin):
    """
    Custom reimplementation of MultiProcPlugin to support UI signaling and GPU queue
    """

    def __init__(self, plugin_args=None):
        # This method implement support for queue signaling
        if "queue" in plugin_args:
            self.queue = plugin_args["queue"]

        super().__init__(plugin_args=plugin_args)

        # it's mandatory delete this argument to avoid plugin copy generated by MapNodes to raise exceptions
        plugin_args["queue"] = None

    def _prerun_check(self, graph):
        """Check if any node exceeds the available resources"""
        # This method implements signaling for insufficient resources error
        try:
            super(MonitoredMultiProcPlugin, self)._prerun_check(graph)
        except RuntimeError:
            self.queue.put(
                WorkflowReport(
                    signal_type=WorkflowSignals.WORKFLOW_INSUFFICIENT_RESOURCES
                )
            )
            raise RuntimeError("Insufficient resources available for job")

    def _report_crash(self, node, result=None):
        # This class implements signaling for generic node error
        try:
            info = None
            for line in result["traceback"]:
                if "out of memory" in line:
                    info = strings.subj_tab_wf_error_oom_gpu
                    break
                elif "Killed" in line:
                    info = strings.subj_tab_wf_error_oom
                    break
                elif "Terminated" in line:
                    info = strings.subj_tab_wf_error_terminated
                    break
            self.queue.put(
                WorkflowReport(
                    long_name=node.fullname,
                    signal_type=WorkflowSignals.NODE_ERROR,
                    info=info,
                )
            )
        except:
            traceback.print_exc()

        return super(MonitoredMultiProcPlugin, self)._report_crash(node, result)

    def _submit_job(self, node, updatehash=False):
        # This class implements signaling for generic node start
        if node.name[0] != "_":
            try:
                self.queue.put(
                    WorkflowReport(
                        long_name=node.fullname,
                        signal_type=WorkflowSignals.NODE_STARTED,
                    )
                )
            except:
                traceback.print_exc()

        # Force english language for every node with: export LC_ALL=en_US.UTF-8
        # This is needed to recognize the "Killed" message in case of Out Of Memory Killer error
        if hasattr(node.interface.inputs, "environ"):
            node.interface.inputs.environ["LC_ALL"] = "en_US.UTF-8"

        return super(MonitoredMultiProcPlugin, self)._submit_job(node, updatehash)

    def _submit_mapnode(self, jobid):
        # This class implements signaling for mapnode start
        try:
            self.queue.put(
                WorkflowReport(
                    long_name=self.procs[jobid].fullname,
                    signal_type=WorkflowSignals.NODE_STARTED,
                )
            )
        except:
            traceback.print_exc()
        return super(MonitoredMultiProcPlugin, self)._submit_mapnode(jobid)

    def _task_finished_cb(self, jobid, cached=False):
        # Implements signaling for generic node completion
        if jobid not in self.mapnodesubids:
            try:
                self.queue.put(
                    WorkflowReport(
                        long_name=self.procs[jobid].fullname,
                        signal_type=WorkflowSignals.NODE_COMPLETED,
                    )
                )
            except:
                traceback.print_exc()
        return super(MonitoredMultiProcPlugin, self)._task_finished_cb(jobid, cached)



# -*- DISCLAIMER: this file contains code derived from Nipype (https://github.com/nipy/nipype/blob/master/LICENSE)  -*-

import shutil
from nipype.interfaces.fsl import BinaryMaths, UnaryMaths, ImageStats, Threshold
from os.path import abspath
import math
from nipype.interfaces.base import (
    BaseInterface,
    BaseInterfaceInputSpec,
    TraitedSpec,
    File,
    isdefined,
)


# nodo per rimozione outliers nel FLAT1
# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterfaceInputSpec)  -*-
class FLAT1OutliersMaskInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="the input image")
    mask_file = File(exists=True, mandatory=True, desc="the original mask image")
    out_file = File(desc="the output mask name")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.TraitedSpec)  -*-
class FLAT1OutliersMaskOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output image")


# -*- DISCLAIMER: this class extends a Nipype class (nipype.interfaces.base.BaseInterface)  -*-
class FLAT1OutliersMask(BaseInterface):
    """
    Creates a mask that can be used to remove the outliers in FLAT1 workflow.

    """

    input_spec = FLAT1OutliersMaskInputSpec
    output_spec = FLAT1OutliersMaskOutputSpec

    def _run_interface(self, runtime):
        self.inputs.out_file = self._gen_outfilename()

        mean = ImageStats()
        mean.inputs.in_file = self.inputs.in_file
        mean.inputs.op_string = "-M"
        mean_res = mean.run()

        mean_value = math.trunc(mean_res.outputs.out_stat)

        if mean_value <= 100:
            threshold = mean_value + 1
            thr = Threshold()
            thr.inputs.in_file = self.inputs.in_file
            thr.inputs.thresh = threshold
            thr_res = thr.run()

            bin_math = UnaryMaths()
            bin_math.inputs.in_file = thr_res.outputs.out_file
            bin_math.inputs.operation = "bin"
            bin_res = bin_math.run()

            sub = BinaryMaths()
            sub.inputs.in_file = self.inputs.mask_file
            sub.inputs.operation = "sub"
            sub.inputs.operand_file = bin_res.outputs.out_file
            sub.inputs.out_file = self.inputs.out_file
            sub.run()
        else:
            shutil.copy(self.inputs.mask_file, self.inputs.out_file)

        return runtime

    def _gen_outfilename(self):
        out_file = self.inputs.out_file
        if not isdefined(out_file):
            out_file = "brain_cortex_mas_refined.nii.gz"
        return abspath(out_file)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self._gen_outfilename()
        return outputs



from PySide6.QtCore import QRunnable, Signal, QObject
from packaging import version
import subprocess
import sys
from swane import __version__
import re


class UpdateCheckSignaler(QObject):
    last_available = Signal(str)


class UpdateCheckWorker(QRunnable):
    """
    Spawn a thread to check swane updates on pip

    """

    def __init__(self):
        super(UpdateCheckWorker, self).__init__()
        self.signal: UpdateCheckSignaler = UpdateCheckSignaler()

    def run(self):

        cmd = sys.executable + " -m pip index versions swane 2>/dev/null"
        output = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE).stdout.decode(
            "utf-8"
        )
        for stdout_line in output.split("\n"):
            regex_pattern = r"^swane \((.+)\)$"
            match = re.match(regex_pattern, stdout_line)
            if match:
                pip_version = match.group(1)
                if UpdateCheckWorker.is_newer_version(pip_version):
                    self.signal.last_available.emit(pip_version)
                    break

    @staticmethod
    def is_newer_version(pip_version: str) -> bool:
        """
            Compare current version with pip version

        Parameters
        ----------
        pip_version : str
            The version retrieved from pip

        Returns
        -------
            True if pip version is newer than current version
        """

        try:
            if version.parse(pip_version) > version.parse(__version__):
                return True
        except:
            pass
        return False



from nipype.interfaces.fsl import (
    BET,
    FLIRT,
    Split,
    ApplyMask,
    ImageStats,
    ImageMaths,
    ApplyXFM,
)
from nipype.interfaces.utility import Merge
from nipype.pipeline.engine import Node
from swane.nipype_pipeline.engine.CustomWorkflow import CustomWorkflow
from swane.nipype_pipeline.nodes.CustomDcm2niix import CustomDcm2niix
from swane.nipype_pipeline.nodes.ForceOrient import ForceOrient
from swane.nipype_pipeline.nodes.VenousCheck import VenousCheck
from nipype.interfaces.utility import IdentityInterface
from configparser import SectionProxy


def venous_workflow(
    name: str,
    venous_dir: str,
    config: SectionProxy,
    venous2_dir: str = None,
    base_dir: str = "/",
) -> CustomWorkflow:
    """
    Analysis of phase contrasts images (in single or two series) to obtain in skull veins
    in reference space, scaled in 0-100 value.

    Parameters
    ----------
    name : str
        The workflow name.
    venous_dir : path
        The directory path of the venous phase contrast DICOM files.
    config: SectionProxy
        workflow settings.
    venous2_dir : path
        If veins phase is divided from anatomic phase, use this param to load the second DICOM files directory.
    base_dir : str, optional
        The base directory path relative to parent workflow. The default is "/".

    Input Node Fields
    ----------
    ref_brain : path
        Betted T13D.

    Returns
    -------
    workflow : CustomWorkflow
        The venous workflow.

    Output Node Fields
    ----------
    veins : path
        Intracranial veins in T13D reference space.

    """

    workflow = CustomWorkflow(name=name, base_dir=base_dir)

    # Input Node
    inputnode = Node(IdentityInterface(fields=["ref_brain"]), name="inputnode")

    # Output Node
    outputnode = Node(IdentityInterface(fields=["veins"]), name="outputnode")

    # NODE 1a: Conversion dicom -> nifti
    veins_conv = Node(CustomDcm2niix(), name="veins_conv")
    veins_conv.inputs.source_dir = venous_dir
    veins_conv.inputs.bids_format = False
    veins_conv.inputs.out_filename = "veins"
    veins_conv.inputs.name_conflicts = 1
    veins_conv.inputs.merge_imgs = 2

    # NODE 2a: Orienting in radiological convention
    veins_reOrient = Node(ForceOrient(), name="veins_reOrient")
    workflow.connect(veins_conv, "converted_files", veins_reOrient, "in_file")

    # NODE 4: Detect the venous phase from the anatomic phase
    veins_check = Node(VenousCheck(), name="veins_check")
    veins_check.long_name = "angiographic volume detection"
    vein_detection_mode = config.getenum_safe("vein_detection_mode")
    veins_check.inputs.detection_mode = vein_detection_mode
    # If the phases are in the same sequence
    if venous2_dir is None:
        # NODE 3a: Divide the two phases from the phase contrast
        veins_split = Node(Split(), name="veins_split")
        veins_split.long_name = "volumes splitting"
        veins_split.inputs.dimension = "t"
        workflow.connect(veins_reOrient, "out_file", veins_split, "in_file")

        workflow.connect(veins_split, "out_files", veins_check, "in_files")
    else:
        # NODE 1b: Conversion dicom -> nifti
        veins2_conv = Node(CustomDcm2niix(), name="veins2_conv")
        veins2_conv.inputs.source_dir = venous2_dir
        veins2_conv.inputs.bids_format = False
        veins2_conv.inputs.out_filename = "veins2"
        veins2_conv.inputs.name_conflicts = 1
        veins2_conv.inputs.merge_imgs = 2

        # NODE 2b: Orienting in radiological convention
        veins2_reOrient = Node(ForceOrient(), name="veins2_reOrient")
        workflow.connect(veins2_conv, "converted_files", veins2_reOrient, "in_file")

        # NODE 3b: Merge the two phases
        veins_merge = Node(Merge(2), name="veins_merge")
        veins_merge.long_name = "volumes merging"
        workflow.connect(veins_reOrient, "out_file", veins_merge, "in1")
        workflow.connect(veins2_reOrient, "out_file", veins_merge, "in2")

        workflow.connect(veins_merge, "out", veins_check, "in_files")

    # NODE 5: Scalp removal and in skull structures segmentation
    bet = Node(BET(), name="veins_bet")
    bet.inputs.mask = True
    bet.inputs.threshold = True
    bet.inputs.surfaces = True
    bet.inputs.frac = config.getfloat_safe("bet_thr")
    workflow.connect(veins_check, "out_file_anat", bet, "in_file")

    # NODE 6: Linear registration of anatomic phase to reference space
    anat_flirt = Node(FLIRT(), name="anat_flirt")
    anat_flirt.long_name = "%s to reference space"
    anat_flirt.inputs.out_matrix_file = "veins2ref.mat"
    anat_flirt.inputs.cost = "mutualinfo"
    anat_flirt.inputs.searchr_x = [-90, 90]
    anat_flirt.inputs.searchr_y = [-90, 90]
    anat_flirt.inputs.searchr_z = [-90, 90]
    anat_flirt.inputs.dof = 6
    anat_flirt.inputs.interp = "trilinear"
    workflow.connect(bet, "out_file", anat_flirt, "in_file")
    workflow.connect(inputnode, "ref_brain", anat_flirt, "reference")

    # NODE 7: Apply in skull mask to venous phase
    veins_inskull_mask = Node(ApplyMask(), name="veins_inskull_mask")
    veins_inskull_mask.long_name = "%s inskull veins"
    workflow.connect(veins_check, "out_file_veins", veins_inskull_mask, "in_file")
    workflow.connect(bet, "inskull_mask_file", veins_inskull_mask, "mask_file")

    # NODE 8: Linear transformation of in skull venous phase in reference space
    veins_flirt = Node(ApplyXFM(), name="veins_flirt")
    veins_flirt.long_name = "%s to reference space"
    veins_flirt.inputs.out_file = "r-veins_inskull.nii.gz"
    veins_flirt.inputs.interp = "trilinear"
    workflow.connect(veins_inskull_mask, "out_file", veins_flirt, "in_file")
    workflow.connect(anat_flirt, "out_matrix_file", veins_flirt, "in_matrix_file")
    workflow.connect(inputnode, "ref_brain", veins_flirt, "reference")

    # NODE 9: Get the max value of venous phase
    veins_range = Node(ImageStats(), name="veins_range")
    veins_range.long_name = "intensity range detection"
    veins_range.inputs.op_string = "-R"
    workflow.connect(veins_flirt, "out_file", veins_range, "in_file")

    # NODE 10: Venous phase rescaling in 0-100
    veins_rescale = Node(ImageMaths(), name="veins_rescale")
    veins_rescale.long_name = "intensity normalization"
    veins_rescale.inputs.out_file = "r-veins_inskull.nii.gz"

    # Function to define the operation string
    def rescale_string(range):
        op_string = "-mul 100 -div %f" % range[1]
        return op_string

    workflow.connect(
        veins_range, ("out_stat", rescale_string), veins_rescale, "op_string"
    )
    workflow.connect(veins_flirt, "out_file", veins_rescale, "in_file")

    workflow.connect(veins_rescale, "out_file", outputnode, "veins")

    return workflow
